"""Set feedback function id to not nullable.

Revision ID: 4
Revises: 3
Create Date: 2024-08-16 12:44:05.560492
"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "4"
down_revision = "3"
branch_labels = None
depends_on = "1"


def upgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###
    with op.batch_alter_table(prefix + "feedbacks") as batch_op:
        batch_op.alter_column(
            "feedback_definition_id",
            existing_type=sa.VARCHAR(length=1024),
            nullable=False,
        )
    # ### end Alembic commands ###


def downgrade(config) -> None:
    prefix = config.get_main_option("trulens.prefix")

    # ### begin Alembic commands ###
    with op.batch_alter_table(prefix + "feedbacks") as batch_op:
        batch_op.alter_column(
            "feedback_definition_id",
            existing_type=sa.VARCHAR(length=1024),
            nullable=True,
        )
    # ### end Alembic commands ###



from datetime import datetime
import logging
from typing import Any, Optional

from opentelemetry.proto.common.v1.common_pb2 import AnyValue
from opentelemetry.proto.common.v1.common_pb2 import ArrayValue
from opentelemetry.proto.common.v1.common_pb2 import KeyValue
from opentelemetry.proto.common.v1.common_pb2 import KeyValueList
from opentelemetry.proto.trace.v1.trace_pb2 import Span as SpanProto
from opentelemetry.proto.trace.v1.trace_pb2 import Status
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.trace import StatusCode
from trulens.core.schema import event as event_schema
from trulens.otel.semconv.trace import SpanAttributes

logger = logging.getLogger(__name__)


def convert_to_any_value(value: Any) -> AnyValue:
    """
    Converts a given value to an AnyValue object.
    This function takes a value of various types (str, bool, int, float, bytes, list, dict)
    and converts it into an AnyValue object. If the value is a list or a dictionary, it
    recursively converts the elements or key-value pairs. Tuples are converted into lists.
    Args:
        value (Any): The value to be converted. It can be of type str, bool, int, float,
                     bytes, list, tuple, or dict.
    Returns:
        AnyValue: The converted AnyValue object.
    Raises:
        ValueError: If the value type is unsupported.
    """
    if isinstance(value, tuple):
        value = list(value)
    any_value = AnyValue()

    if isinstance(value, str):
        any_value.string_value = value
    elif isinstance(value, bool):
        any_value.bool_value = value
    elif isinstance(value, int):
        any_value.int_value = value
    elif isinstance(value, float):
        any_value.double_value = value
    elif isinstance(value, bytes):
        any_value.bytes_value = value
    elif isinstance(value, list):
        array_value = ArrayValue()
        for item in value:
            array_value.values.append(convert_to_any_value(item))
        any_value.array_value.CopyFrom(array_value)
    elif isinstance(value, dict):
        kv_list = KeyValueList()
        for k, v in value.items():
            kv = KeyValue(key=k, value=convert_to_any_value(v))
            kv_list.values.append(kv)
        any_value.kvlist_value.CopyFrom(kv_list)
    else:
        raise ValueError(f"Unsupported value type: {type(value)}")

    return any_value


def convert_readable_span_to_proto(span: ReadableSpan) -> SpanProto:
    """
    Converts a ReadableSpan object to a protobuf object for a Span.
    Args:
        span (ReadableSpan): The span to be converted.
    Returns:
        SpanProto: The converted span in SpanProto format.
    """
    span_proto = SpanProto(
        trace_id=span.context.trace_id.to_bytes(16, "big")
        if span.context
        else b"",
        span_id=span.context.span_id.to_bytes(8, "big")
        if span.context
        else b"",
        parent_span_id=span.parent.span_id.to_bytes(8, "big")
        if span.parent
        else b"",
        name=span.name,
        kind=SpanProto.SpanKind.SPAN_KIND_INTERNAL,
        start_time_unix_nano=span.start_time if span.start_time else 0,
        end_time_unix_nano=span.end_time if span.end_time else 0,
        attributes=[
            KeyValue(key=k, value=convert_to_any_value(v))
            for k, v in span.attributes.items()
        ]
        if span.attributes
        else None,
        status=Status(code=Status.StatusCode.STATUS_CODE_UNSET),
    )
    return span_proto


def to_timestamp(timestamp: Optional[int]) -> datetime:
    """
    Utility function for converting OTEL timestamps to datetime objects.
    """
    if timestamp:
        return datetime.fromtimestamp(timestamp * 1e-9)

    return datetime.now()


def check_if_trulens_span(span: ReadableSpan) -> bool:
    """
    Check if a given span is a TruLens span.
    This function checks the attributes of the provided span to determine if it
    contains a TruLens-specific attribute, identified by the presence of
    `SpanAttributes.RECORD_ID`.
    Args:
        span (ReadableSpan): The span to be checked.
    Returns:
        bool: True if the span contains the TruLens-specific attribute, False otherwise.
    """
    if not span.attributes:
        return False
    app_name = span.attributes.get(SpanAttributes.APP_NAME)
    return bool(app_name)


def construct_event(span: ReadableSpan) -> event_schema.Event:
    context = span.get_span_context()
    parent = span.parent

    if context is None:
        raise ValueError("Span context is None")

    return event_schema.Event(
        event_id=str(context.span_id),
        record={
            "name": span.name,
            "kind": SpanProto.SpanKind.SPAN_KIND_INTERNAL,
            "parent_span_id": str(parent.span_id if parent else ""),
            "status": "STATUS_CODE_ERROR"
            if span.status.status_code == StatusCode.ERROR
            else "STATUS_CODE_UNSET",
        },
        record_attributes=span.attributes,
        record_type=event_schema.EventRecordType.SPAN,
        resource_attributes=span.resource.attributes,
        start_timestamp=to_timestamp(span.start_time),
        timestamp=to_timestamp(span.end_time),
        trace={
            "span_id": str(context.span_id),
            "trace_id": str(context.trace_id),
            "parent_id": str(parent.span_id if parent else ""),
        },
    )



"""Debugging utilities."""

import asyncio
import io
import threading
from typing import Optional

from trulens.core.utils import python as python_utils
from trulens.core.utils import text as text_utils


def iprint(*args, tabs=0, **kwargs):
    """Print with thread and task info,"""

    thread = threading.current_thread()
    thread_ident = thread.name

    try:
        task = asyncio.current_task()
        task_ident = task.get_name()
    except RuntimeError:
        task_ident = "no running loop"

    textbuffer = io.StringIO()

    print(*args, **kwargs, file=textbuffer, end="")

    text = textbuffer.getvalue()

    tabbed = text_utils.retab(
        text, tab=f"[{thread_ident}][{task_ident}]" + ("    " * tabs) + " "
    )

    print(tabbed)


def print_context(msg: Optional[str] = None, tabs=0):
    """Print the status of trulens context variables."""

    from trulens.core import instruments as core_instruments
    from trulens.core.feedback import endpoint as core_endpoint

    if msg is None:
        msg = python_utils.code_line(
            python_utils.caller_frameinfo(offset=1), show_source=True
        )

    contextvars = [
        core_instruments.WithInstrumentCallbacks._context_contexts,
        core_instruments.WithInstrumentCallbacks._stack_contexts,
        core_endpoint.Endpoint._context_endpoints,
    ]

    iprint(msg, tabs=tabs)

    for contextvar in contextvars:
        iprint(f"{contextvar.name} [{len(contextvar.get())}]", tabs=tabs + 1)



from __future__ import annotations

from abc import ABC
from abc import ABCMeta
from abc import abstractmethod
import contextvars
import datetime
import inspect
from inspect import BoundArguments
from inspect import Signature
import json
import logging
import os
import threading
from typing import (
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Dict,
    Hashable,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)
import weakref

import pandas as pd
import pydantic
from trulens.core import experimental as core_experimental
from trulens.core import instruments as core_instruments
from trulens.core import session as core_session
from trulens.core._utils import optional as optional_utils
from trulens.core._utils.pycompat import Future  # import standard exception
from trulens.core.database import base as core_db
from trulens.core.database import connector as core_connector
from trulens.core.feedback import endpoint as core_endpoint
from trulens.core.feedback import feedback as core_feedback
from trulens.core.run import Run
from trulens.core.run import RunConfig
from trulens.core.run import validate_dataset_spec
from trulens.core.schema import app as app_schema
from trulens.core.schema import base as base_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import record as record_schema
from trulens.core.schema import select as select_schema
from trulens.core.session import TruSession
from trulens.core.utils import asynchro as asynchro_utils
from trulens.core.utils import constants as constant_utils
from trulens.core.utils import containers as container_utils
from trulens.core.utils import deprecation as deprecation_utils
from trulens.core.utils import imports as import_utils
from trulens.core.utils import json as json_utils
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import signature as signature_utils
from trulens.core.utils import threading as threading_utils
from trulens.otel.semconv.constants import (
    TRULENS_RECORD_ROOT_INSTRUMENT_WRAPPER_FLAG,
)
from trulens.otel.semconv.trace import SpanAttributes

logger = logging.getLogger(__name__)

# App component.
COMPONENT = Any

A = TypeVar("A")
T = TypeVar("T")

# Message produced when an attribute is looked up from our App but is actually
# an attribute of the enclosed app.
ATTRIBUTE_ERROR_MESSAGE = """
{class_name} has no attribute `{attribute_name}` but the wrapped app {app_class_name} does. If
you are calling a {app_class_name} method, retrieve it from that app instead of from
{class_name}. If you need to record your app's behavior, use {class_name} as a context
manager as in this Example:
```python
    app: {app_class_name} = ...  # your app
    truapp: {class_name} = {class_name}(app, ...)  # the truera recorder

    with truapp as recorder:
      result = app.{attribute_name}(...)

    record: Record = recorder.get() # get the record of the invocation if needed
```
"""

_component_impls: Dict[str, Type[ComponentView]] = {}


class ComponentViewMeta(ABCMeta):
    def __init__(
        cls,
        classname: str,
        bases: Tuple[Type[ComponentView]],
        dict_: Dict[str, Any],
    ):
        newtype = type.__init__(cls, classname, bases, dict_)

        if hasattr(cls, "component_of_json"):
            # Only register the higher-level classes that can enumerate this
            # subclasses. Otherwise an infinite loop will result in the of_json
            # logic below.

            _component_impls[classname] = cls  # type: ignore[assignment]
        return newtype


class ComponentView(ABC, metaclass=ComponentViewMeta):
    """
    Views of common app component types for sorting them and displaying them in
    some unified manner in the UI. Operates on components serialized into json
    dicts representing various components, not the components themselves.
    """

    def __init__(self, json: serial_utils.JSON):
        self.json = json
        self.cls = pyschema_utils.Class.of_class_info(json)

    @classmethod
    def of_json(cls, json: serial_utils.JSON) -> "ComponentView":
        """Sort the given json into the appropriate component view type."""

        cls_obj = pyschema_utils.Class.of_class_info(json)

        for _, view in _component_impls.items():
            # NOTE: includes prompt, llm, tool, agent, memory, other which may be overridden
            if view.class_is(cls_obj):
                return view.of_json(json)

        raise TypeError(f"Unhandled component type with class {cls_obj}")

    @staticmethod
    @abstractmethod
    def class_is(cls_obj: pyschema_utils.Class) -> bool:
        """Determine whether the given class representation `cls` is of the type to
        be viewed as this component type."""
        pass

    def unsorted_parameters(
        self, skip: Set[str]
    ) -> Dict[str, serial_utils.JSON_BASES_T]:
        """All basic parameters not organized by other accessors."""

        ret = {}

        for k, v in self.json.items():
            if k not in skip and isinstance(v, serial_utils.JSON_BASES):
                ret[k] = v

        return ret

    @staticmethod
    def innermost_base(
        bases: Optional[Sequence[pyschema_utils.Class]] = None,
        among_modules=set(["langchain", "llama_index", "trulens"]),
    ) -> Optional[str]:
        """
        Given a sequence of classes, return the first one which comes from one
        of the `among_modules`. You can use this to determine where ultimately
        the encoded class comes from in terms of langchain, llama_index, or
        trulens even in cases they extend each other's classes. Returns
        None if no module from `among_modules` is named in `bases`.
        """
        if bases is None:
            return None

        for base in bases:
            if "." in base.module.module_name:
                root_module = base.module.module_name.split(".")[0]
            else:
                root_module = base.module.module_name

            if root_module in among_modules:
                return root_module

        return None


class TrulensComponent(ComponentView):
    """Components provided in trulens."""

    @staticmethod
    def class_is(cls_obj: pyschema_utils.Class) -> bool:
        if ComponentView.innermost_base(cls_obj.bases) == "trulens":
            return True

        # if any(base.module.module_name.startswith("trulens.") for base in cls.bases):
        #    return True

        return False

    @staticmethod
    def of_json(json: serial_utils.JSON) -> "TrulensComponent":
        # NOTE: This import is here to avoid circular imports.
        from trulens.core.utils.trulens import component_of_json

        return component_of_json(json)


class Prompt(ComponentView):
    # langchain.prompts.base.BasePromptTemplate
    # llama_index.prompts.base.Prompt

    @property
    @abstractmethod
    def template(self) -> str:
        pass


class LLM(ComponentView):
    # langchain.llms.base.BaseLLM
    # llama_index.llms.base.LLM

    @property
    @abstractmethod
    def model_name(self) -> str:
        pass


class Tool(ComponentView):
    # langchain ???
    # llama_index.tools.types.BaseTool

    @property
    @abstractmethod
    def tool_name(self) -> str:
        pass


class Agent(ComponentView):
    # langchain ???
    # llama_index.agent.types.BaseAgent

    @property
    @abstractmethod
    def agent_name(self) -> str:
        pass


class Memory(ComponentView):
    # langchain.schema.BaseMemory
    # llama_index ???
    pass


class Other(ComponentView):
    # Any component that does not fit into the other named categories.
    pass


class CustomComponent(ComponentView):
    class Custom(Other):
        # No categorization of custom class components for now. Using just one
        # "Custom" catch-all.

        @staticmethod
        def class_is(cls_obj: pyschema_utils.Class) -> bool:
            return True

    COMPONENT_VIEWS = [Custom]

    @staticmethod
    def constructor_of_class(
        cls_obj: pyschema_utils.Class,
    ) -> Type["CustomComponent"]:
        for view in CustomComponent.COMPONENT_VIEWS:
            if view.class_is(cls_obj):
                return view

        raise TypeError(f"Unknown custom component type with class {cls_obj}")

    @staticmethod
    def component_of_json(json: serial_utils.JSON) -> "CustomComponent":
        cls = pyschema_utils.Class.of_class_info(json)

        view = CustomComponent.constructor_of_class(cls)

        return view(json)

    @staticmethod
    def class_is(cls_obj: pyschema_utils.Class) -> bool:
        # Assumes this is the last check done.
        return True

    @classmethod
    def of_json(cls, json: serial_utils.JSON) -> "CustomComponent":
        return CustomComponent.component_of_json(json)


def instrumented_component_views(
    obj: object,
) -> Iterable[Tuple[serial_utils.Lens, ComponentView]]:
    """
    Iterate over contents of `obj` that are annotated with the CLASS_INFO
    attribute/key. Returns triples with the accessor/selector, the Class object
    instantiated from CLASS_INFO, and the annotated object itself.
    """

    for q, o in serial_utils.all_objects(obj):
        if (
            isinstance(o, pydantic.BaseModel)
            and constant_utils.CLASS_INFO in o.model_fields
        ):
            yield q, ComponentView.of_json(json=o)

        if isinstance(o, Dict) and constant_utils.CLASS_INFO in o:
            yield q, ComponentView.of_json(json=o)


def _can_import(to_import: str) -> bool:
    try:
        __import__(to_import)
        return True
    except ImportError:
        return False


class App(
    app_schema.AppDefinition,
    core_instruments.WithInstrumentCallbacks,
    Hashable,
):
    """Base app recorder type.

    Non-serialized fields here while the serialized ones are defined in
    [AppDefinition][trulens.core.schema.app.AppDefinition].

    This class is abstract. Use one of these concrete subclasses as appropriate:
    - [TruLlama][trulens.apps.llamaindex.TruLlama] for _LlamaIndex_ apps.
    - [TruChain][trulens.apps.langchain.TruChain] for _LangChain_ apps.
    - [TruRails][trulens.apps.nemo.TruRails] for _NeMo Guardrails_
        apps.
    - [TruVirtual][trulens.apps.virtual.TruVirtual] for recording
        information about invocations of apps without access to those apps.
    - [TruCustomApp][trulens.apps.custom.TruCustomApp] (To be deprecated in favor of TruApp) for custom
        apps. These need to be decorated to have appropriate data recorded.
    - [TruApp][trulens.apps.app.TruApp] for custom
        apps allowing maximized flexibility. These need to be decorated to have appropriate data recorded.
    - [TruBasicApp][trulens.apps.basic.TruBasicApp] for apps defined
        solely by a string-to-string method.
    """

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        # Tru, DB, most of the types on the excluded fields.
        arbitrary_types_allowed=True
    )

    feedbacks: List[core_feedback.Feedback] = pydantic.Field(
        exclude=True, default_factory=list
    )
    """Feedback functions to evaluate on each record."""

    session: core_session.TruSession = pydantic.Field(
        default_factory=core_session.TruSession, exclude=True
    )
    """Session for this app."""

    @property
    def connector(self) -> core_connector.DBConnector:
        """Database connector."""

        return self.session.connector

    @property
    def db(self) -> core_db.DB:
        """Database used by this app."""

        return self.connector.db

    @deprecation_utils.deprecated_property(
        "The `App.tru` property for retrieving `Tru` is deprecated. "
        "Use `App.connector` which contains the replacement `core_connector.DBConnector` class instead."
    )
    def tru(self) -> core_connector.DBConnector:
        return self.connector

    app: Any = pydantic.Field(exclude=True)
    """The app to be recorded."""

    main_method_name: Optional[str] = pydantic.Field(None)
    """Name of the main method of the app to be recorded. For serialization and this is required for OTEL."""

    instrument: Optional[core_instruments.Instrument] = pydantic.Field(
        None, exclude=True
    )
    """Instrumentation class.

    This is needed for serialization as it tells us which objects we want to be
    included in the json representation of this app.
    """

    recording_contexts: contextvars.ContextVar[
        core_instruments._RecordingContext
    ] = pydantic.Field(None, exclude=True)
    """Sequences of records produced by the this class used as a context manager
    are stored in a RecordingContext.

    Using a context var so that context managers can be nested.
    """

    instrumented_methods: Dict[int, Dict[Callable, serial_utils.Lens]] = (
        pydantic.Field(exclude=True, default_factory=dict)
    )
    """Mapping of instrumented methods (by id(.) of owner object and the
    function) to their path in this app."""

    records_with_pending_feedback_results: container_utils.BlockingSet[
        record_schema.Record
    ] = pydantic.Field(
        exclude=True, default_factory=container_utils.BlockingSet
    )
    """Records produced by this app which might have yet to finish
    feedback runs."""

    manage_pending_feedback_results_thread: Optional[threading_utils.Thread] = (
        pydantic.Field(exclude=True, default=None)
    )
    """Thread for manager of pending feedback results queue.

    See _manage_pending_feedback_results."""

    selector_check_warning: bool = False
    """Issue warnings when selectors are not found in the app with a placeholder
    record.

    If False, constructor will raise an error instead.
    """

    selector_nocheck: bool = False
    """Ignore selector checks entirely.

    This may be necessary 1if the expected record content cannot be determined
    before it is produced.
    """

    _context_vars_tokens: Dict[contextvars.ContextVar, contextvars.Token] = (
        pydantic.PrivateAttr(default_factory=dict)
    )

    snowflake_object_type: Optional[str] = pydantic.Field(
        None,
    )
    snowflake_object_name: Optional[str] = pydantic.Field(
        None,
    )
    snowflake_object_version: Optional[str] = pydantic.Field(
        None,
    )

    snowflake_run_dao: Optional[Any] = pydantic.Field(None, exclude=True)

    snowflake_app_dao: Optional[Any] = pydantic.Field(None, exclude=True)

    def __init__(
        self,
        connector: Optional[core_connector.DBConnector] = None,
        feedbacks: Optional[Iterable[core_feedback.Feedback]] = None,
        **kwargs,
    ):
        if feedbacks is not None:
            feedbacks = list(feedbacks)
        else:
            feedbacks = []

        # for us:
        if connector:
            tru_session = TruSession(connector=connector)
            if tru_session.connector is not connector:
                raise ValueError(
                    "Already created `TruSession` with different `connector`!"
                )
            kwargs["connector"] = connector

        kwargs["feedbacks"] = feedbacks
        kwargs["recording_contexts"] = contextvars.ContextVar(
            "recording_contexts", default=None
        )
        app = kwargs["app"]

        otel_enabled = TruSession().experimental_feature(
            core_experimental.Feature.OTEL_TRACING
        )
        main_method = None

        if otel_enabled:
            if app is None:
                raise ValueError(
                    "A valid app instance must be provided when specifying 'main_method'."
                )

            if "main_method" in kwargs:
                main_method = kwargs["main_method"]

                # Instead of always checking for binding,  enforce it except when app is an instance of TruWrapperApp (tru basic app).
                try:
                    from trulens.apps.basic import TruWrapperApp
                except ImportError:
                    TruWrapperApp = None

                if TruWrapperApp is None or not isinstance(app, TruWrapperApp):
                    if (
                        not hasattr(main_method, "__self__")
                        or main_method.__self__ != app
                    ):
                        raise ValueError(
                            f"main_method `{main_method.__name__}` must be bound to the provided `app` instance."
                        )

                self._wrap_main_function(app, main_method.__name__)

        super().__init__(**kwargs)

        if main_method:
            self.main_method_name = main_method.__name__  # for serialization

        self._current_context_manager_lock = threading.Lock()
        self._current_context_manager = None

        if connector and _can_import("trulens.connectors.snowflake"):
            from trulens.connectors.snowflake import SnowflakeConnector
            from trulens.connectors.snowflake.dao.enums import ObjectType

            if isinstance(connector, SnowflakeConnector):
                self.snowflake_object_type = (
                    ObjectType.EXTERNAL_AGENT.value
                    if "object_type" not in kwargs
                    or kwargs["object_type"] is None
                    else kwargs["object_type"]
                )

                (
                    self.snowflake_app_dao,
                    self.snowflake_run_dao,
                    self.snowflake_object_name,
                    self.snowflake_object_version,
                ) = connector.initialize_snowflake_dao_fields(
                    object_type=self.snowflake_object_type,
                    app_name=kwargs["app_name"],
                    app_version=kwargs["app_version"],
                )

        self.app = app

        if self.instrument is not None:
            self.instrument.instrument_object(
                obj=self.app, query=select_schema.Select.Query().app
            )
        else:
            pass

        if self.feedback_mode == feedback_schema.FeedbackMode.WITH_APP_THREAD:
            self._start_manage_pending_feedback_results()

        self._tru_post_init()

    def __del__(self):
        """Shut down anything associated with this app that might persist otherwise."""
        try:
            # Use object.__getattribute__ to avoid triggering __getattr__
            m_thread = object.__getattribute__(
                self, "manage_pending_feedback_results_thread"
            )
        except Exception:
            m_thread = None

        if m_thread is not None:
            try:
                records = object.__getattribute__(
                    self, "records_with_pending_feedback_results"
                )
                if records is not None:
                    records.shutdown()
            except Exception:
                # If records or shutdown is not available, ignore.
                pass

            try:
                m_thread.join()
            except Exception:
                pass

    @staticmethod
    def _has_record_root_instrumentation(func: Callable) -> bool:
        while hasattr(func, "__wrapped__"):
            if hasattr(func, TRULENS_RECORD_ROOT_INSTRUMENT_WRAPPER_FLAG):
                return True
            func = func.__wrapped__
        return False

    def _wrap_main_function(self, app: Any, method_name: str) -> None:
        if TruSession().experimental_feature(
            core_experimental.Feature.OTEL_TRACING, freeze=True
        ):
            from trulens.core.otel.instrument import instrument

            if not hasattr(app, method_name):
                raise ValueError(f"App must have an `{method_name}` method!")
            func = getattr(app, method_name)
            if self._has_record_root_instrumentation(func):
                return
            sig = inspect.signature(func)
            wrapper = instrument(
                span_type=SpanAttributes.SpanType.RECORD_ROOT,
                attributes=lambda ret, exception, *args, **kwargs: {
                    # langchain has specific main input/output logic.
                    SpanAttributes.RECORD_ROOT.INPUT: self.main_input(
                        func, sig, sig.bind_partial(**kwargs)
                    ),
                    SpanAttributes.RECORD_ROOT.OUTPUT: self.main_output(
                        func, sig, sig.bind_partial(**kwargs), ret
                    ),
                },
                is_app_specific_record_root=True,
            )
            # HACK!: This is a major hack to get around the fact that we can't
            # set the desired method on the app object due to Pydantic only
            # allowing fields to be set on the class, not on the instance for
            # some reason. To get around this, we're setting it on the __dict__
            # of the app object, which is mutable and is the first place that
            # the field is looked up it seems. There's another implication of
            # this, which is that the desired method for this object will not
            # run whatever is instrumented by TruChain otherwise but that's
            # fine.
            app.__dict__[method_name] = wrapper(func)

    def _start_manage_pending_feedback_results(self) -> None:
        """Start the thread that manages the queue of records with
        pending feedback results.

        This is meant to be run permanently in a separate thread. It will
        remove records from the set `records_with_pending_feedback_results` as
        their feedback results are computed.
        """

        if self.manage_pending_feedback_results_thread is not None:
            raise RuntimeError("Manager Thread already started.")

        self.manage_pending_feedback_results_thread = threading.Thread(
            target=self._manage_pending_feedback_results,
            args=(weakref.proxy(self),),
            daemon=True,  # otherwise this thread will keep parent alive
            name=f"manage_pending_feedback_results_thread(app_name={self.app_name}, app_version={self.app_version})",
        )
        self.manage_pending_feedback_results_thread.start()

    @staticmethod
    def _manage_pending_feedback_results(
        self_proxy: weakref.ProxyType[App],
    ) -> None:
        """Manage the queue of records with pending feedback results.

        This is meant to be run permanently in a separate thread. It will
        remove records from the set records_with_pending_feedback_results as
        their feedback results are computed.
        """

        try:
            while True:
                record = self_proxy.records_with_pending_feedback_results.pop()
                record.wait_for_feedback_results()

        except StopIteration:
            pass
            # Set has been shut down.
        except ReferenceError:
            pass
            # self was unloaded, shut down as well.

    def wait_for_feedback_results(
        self, feedback_timeout: Optional[float] = None
    ) -> Iterable[record_schema.Record]:
        """Wait for all feedbacks functions to complete.

        Args:
            feedback_timeout: Timeout in seconds for waiting for feedback
                results for each feedback function. Note that this is not the
                total timeout for this entire blocking call.

        Returns:
            An iterable of records that have been waited on. Note a record will be
                included even if a feedback computation for it failed or
                timed out.

        This applies to all feedbacks on all records produced by this app. This
        call will block until finished and if new records are produced while
        this is running, it will include them.
        """

        while (
            record := self.records_with_pending_feedback_results.pop(
                blocking=False
            )
        ) is not None:
            record.wait_for_feedback_results(feedback_timeout=feedback_timeout)
            yield record

    @classmethod
    def select_context(cls, app: Optional[Any] = None) -> serial_utils.Lens:
        """Try to find retriever components in the given `app` and return a lens to
        access the retrieved contexts that would appear in a record were these
        components to execute."""

        # Catch the old case where a user calls App.select_context and gives
        # their app (not an App instance) as the app arg.
        if app is not None:
            mod = app.__class__.__module__
            if mod.startswith("langchain"):
                with import_utils.OptionalImports(
                    messages=optional_utils.REQUIREMENT_APPS_LANGCHAIN
                ):
                    from trulens.apps.langchain.tru_chain import TruChain

                return TruChain.select_context(app=app)
            elif mod.startswith("llama_index"):
                with import_utils.OptionalImports(
                    messages=optional_utils.REQUIREMENT_APPS_LLAMA
                ):
                    from trulens.apps.llamaindex.tru_llama import TruLlama

                return TruLlama.select_context(app=app)
            elif mod.startswith("nemoguardrails"):
                with import_utils.OptionalImports(
                    messages=optional_utils.REQUIREMENT_APPS_NEMO
                ):
                    from trulens.apps.nemo.tru_rails import TruRails

                return TruRails.select_context(app=app)
            else:
                raise ValueError(
                    f"Cannot determine the app type from its module {mod}."
                )

        raise NotImplementedError(
            f"`select_context` not implemented for {cls.__name__}. "
            "Call `select_context` using the appropriate subclass (TruChain, TruLlama, TruRails, etc)."
        )

    def __hash__(self):
        return hash(id(self))

    def _tru_post_init(self):
        """
        Database-related initialization and additional data checks.

        DB:
            - Insert the app into the database.
            - Insert feedback function definitions into the database.

        Checks:
            - In deferred mode, try to serialize and deserialize feedback functions.
            - Check that feedback function selectors are likely to refer to expected
                app or record components.

        """

        if self.connector is None:
            if self.feedback_mode != feedback_schema.FeedbackMode.NONE:
                logger.debug("Using default database connector.")
                self.connector = core_connector.DefaultDBConnector()

        else:
            if self.feedback_mode == feedback_schema.FeedbackMode.NONE:
                logger.warning(
                    "`connector` is specified but `feedback_mode` is `FeedbackMode.NONE`. "
                    "No feedback evaluation and logging will occur."
                )

        otel_tracing_enabled = os.getenv(
            "TRULENS_OTEL_TRACING", ""
        ).lower() in ["1", "true"]

        if self.connector is not None and not otel_tracing_enabled:
            self.connector.add_app(app=self)

            if self.feedback_mode != feedback_schema.FeedbackMode.NONE:
                logger.debug("Inserting feedback function definitions to db.")

                for f in self.feedbacks:
                    self.connector.add_feedback_definition(f)

        else:
            if len(self.feedbacks) > 0:
                raise ValueError(
                    "Feedback logging requires `App.connector` to be specified."
                )

        for f in self.feedbacks:
            if (
                self.feedback_mode == feedback_schema.FeedbackMode.DEFERRED
                or f.run_location
                == feedback_schema.FeedbackRunLocation.SNOWFLAKE
            ):
                if (
                    isinstance(f.implementation, pyschema_utils.Method)
                    and f.implementation.obj.cls.module.module_name
                    == "trulens.providers.cortex.provider"
                    and f.implementation.obj.cls.name == "Cortex"
                ):
                    continue
                # Try to load each of the feedback implementations. Deferred
                # mode will do this but we want to fail earlier at app
                # constructor here.
                try:
                    f.implementation.load()
                except Exception as e:
                    raise Exception(
                        f"Feedback function {f} is not loadable. Cannot use DEFERRED feedback mode. {e}"
                    ) from e

        if not self.selector_nocheck:
            dummy = self.dummy_record()

            for feedback in self.feedbacks:
                feedback.check_selectors(
                    app=self,
                    # Don't have a record yet, but use an empty one for the non-call related fields.
                    record=dummy,
                    warning=self.selector_check_warning,
                )

    def main_call(self, human: str) -> str:
        """If available, a single text to a single text invocation of this app."""

        if self.__class__.main_acall is not App.main_acall:
            # Use the async version if available.
            return asynchro_utils.sync(self.main_acall, human)

        raise NotImplementedError()

    async def main_acall(self, human: str) -> str:
        """If available, a single text to a single text invocation of this app."""

        if self.__class__.main_call is not App.main_call:
            logger.warning("Using synchronous version of main call.")
            # Use the sync version if available.
            return await asynchro_utils.desync(self.main_call, human)

        raise NotImplementedError()

    def _extract_content(self, value, content_keys=["content"]):
        """
        Extracts the 'content' from various data types commonly used by libraries
        like OpenAI, Canopy, LiteLLM, etc. This method navigates nested data
        structures (pydantic models, dictionaries, lists) to retrieve the
        'content' field. If 'content' is not directly available, it attempts to
        extract from known structures like 'choices' in a ChatResponse. This
        standardizes extracting relevant text or data from complex API responses
        or internal data representations.

        Args:
            value: The input data to extract content from. Can be a pydantic
                   model, dictionary, list, or basic data type.

        Returns:
            The extracted content, which may be a single value, a list of values,
            or a nested structure with content extracted from all levels.
        """
        return signature_utils._extract_content(
            value, content_keys=content_keys
        )

    def main_input(
        self, func: Callable, sig: Signature, bindings: BoundArguments
    ) -> str:
        """Determine (guess) the main input string for a main app call.

        Args:
            func: The main function we are targeting in this determination.

            sig: The signature of the above.

            bindings: The arguments to be passed to the function.

        Returns:
            The main input string.
        """
        return signature_utils.main_input(func, sig, bindings)

    def main_output(
        self,
        func: Callable,  # pylint: disable=W0613
        sig: Signature,  # pylint: disable=W0613
        bindings: BoundArguments,  # pylint: disable=W0613
        ret: Any,
    ) -> str:
        return signature_utils.main_output(func, ret)

    # WithInstrumentCallbacks requirement
    def on_method_instrumented(
        self, obj: object, func: Callable, path: serial_utils.Lens
    ):
        """Called by instrumentation system for every function requested to be
        instrumented by this app."""

        if id(obj) in self.instrumented_methods:
            funcs = self.instrumented_methods[id(obj)]

            if func in funcs:
                old_path = funcs[func]

                if path != old_path:
                    logger.warning(
                        "Method %s was already instrumented on path %s. "
                        "Calls at %s may not be recorded.",
                        func,
                        old_path,
                        path,
                    )

                return

            else:
                funcs[func] = path

        else:
            funcs = dict()
            self.instrumented_methods[id(obj)] = funcs
            funcs[func] = path

    # WithInstrumentCallbacks requirement
    def get_methods_for_func(
        self, func: Callable
    ) -> Iterable[Tuple[int, Callable, serial_utils.Lens]]:
        """Get the methods (rather the inner functions) matching the given
        `func` and the path of each.

        See
        [WithInstrumentCallbacks.get_methods_for_func][trulens.core.instruments.WithInstrumentCallbacks.get_methods_for_func].
        """

        for _id, funcs in self.instrumented_methods.items():
            for f, path in funcs.items():
                if f == func:
                    yield (_id, f, path)

    # WithInstrumentCallbacks requirement
    def get_method_path(self, obj: object, func: Callable) -> serial_utils.Lens:
        """Get the path of the instrumented function `method` relative to this
        app."""

        # TODO: cleanup and/or figure out why references to objects change when executing langchain chains.

        funcs = self.instrumented_methods.get(id(obj))

        if funcs is None:
            logger.warning(
                "A new object of type %s at %s is calling an instrumented method %s. "
                "The path of this call may be incorrect.",
                python_utils.class_name(type(obj)),
                python_utils.id_str(obj),
                python_utils.callable_name(func),
            )
            try:
                _id, _, path = next(iter(self.get_methods_for_func(func)))

            except Exception:
                logger.warning(
                    "No other objects use this function so cannot guess path."
                )
                return None

            logger.warning(
                "Guessing path of new object is %s based on other object (%s) using this function.",
                path,
                python_utils.id_str(_id),
            )

            funcs = {func: path}

            self.instrumented_methods[id(obj)] = funcs

            return path

        else:
            if func not in funcs:
                logger.warning(
                    "A new object of type %s at %s is calling an instrumented method %s. "
                    "The path of this call may be incorrect.",
                    python_utils.class_name(type(obj)),
                    python_utils.id_str(obj),
                    python_utils.callable_name(func),
                )

                try:
                    _id, _, path = next(iter(self.get_methods_for_func(func)))
                except Exception:
                    logger.warning(
                        "No other objects use this function so cannot guess path."
                    )
                    return None

                logger.warning(
                    "Guessing path of new object is %s based on other object (%s) using this function.",
                    path,
                    python_utils.id_str(_id),
                )

                return path

            else:
                return funcs.get(func)

    def json(self, *args, **kwargs):
        """Create a json string representation of this app."""
        # Need custom jsonification here because it is likely the model
        # structure contains loops.

        return json_utils.json_str_of_obj(
            self, *args, instrument=self.instrument, **kwargs
        )

    def model_dump(self, *args, redact_keys: bool = False, **kwargs):
        # Same problem as in json.
        return json_utils.jsonify(
            self,
            instrument=self.instrument,
            redact_keys=redact_keys,
            *args,
            **kwargs,
        )

    def _prevent_invalid_otel_syntax(self):
        if self.session.experimental_feature(
            core_experimental.Feature.OTEL_TRACING
        ):
            raise RuntimeError("Invalid TruLens OTEL Tracing syntax.")

    # For use as a context manager.
    def __enter__(self):
        if self.session.experimental_feature(
            core_experimental.Feature.OTEL_TRACING
        ):
            from trulens.core.otel.instrument import OtelRecordingContext

            with self._current_context_manager_lock:
                if self._current_context_manager is not None:
                    raise RuntimeError(
                        "Already recording with a context manager, cannot nest!"
                    )
                self._current_context_manager = OtelRecordingContext(
                    app_name=self.app_name,
                    app_version=self.app_version,
                    run_name="",
                    input_id="",
                )
            return self._current_context_manager.__enter__()

        if not core_instruments.Instrument._have_context():
            raise RuntimeError(core_endpoint._NO_CONTEXT_WARNING)

        self._prevent_invalid_otel_syntax()
        ctx = core_instruments._RecordingContext(app=self)

        token = self.recording_contexts.set(ctx)
        ctx.token = token

        return ctx

    # For use as a context manager.
    def __exit__(self, exc_type, exc_value, exc_tb):
        if self.session.experimental_feature(
            core_experimental.Feature.OTEL_TRACING
        ):
            with self._current_context_manager_lock:
                if self._current_context_manager is None:
                    raise RuntimeError("Unknown recording context manager!")
                context_manager = self._current_context_manager
            return context_manager.__exit__(exc_type, exc_value, exc_tb)

        self._prevent_invalid_otel_syntax()

        ctx = self.recording_contexts.get()
        self.recording_contexts.reset(ctx.token)

        if exc_type is not None:
            raise exc_value

        return

    # For use as a context manager.
    async def __aenter__(self):
        self._prevent_invalid_otel_syntax()

        ctx = core_instruments._RecordingContext(app=self)

        token = self.recording_contexts.set(ctx)
        ctx.token = token

        # self._set_context_vars()

        return ctx

    # For use as a context manager.
    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self._prevent_invalid_otel_syntax()

        ctx = self.recording_contexts.get()
        self.recording_contexts.reset(ctx.token)

        # self._reset_context_vars()

        if exc_type is not None:
            raise exc_value

        return

    def _set_context_vars(self):
        # HACK: For debugging purposes, try setting/resetting all context vars
        # used in trulens around the app context managers due to bugs in trying
        # to set/reset them where more appropriate. This is not ideal as not
        # resetting context vars where appropriate will result possibly in
        # incorrect tracing information.

        from trulens.core.feedback.endpoint import Endpoint
        from trulens.core.instruments import WithInstrumentCallbacks

        CONTEXT_VARS = [
            WithInstrumentCallbacks._stack_contexts,
            WithInstrumentCallbacks._context_contexts,
            Endpoint._context_endpoints,
        ]

        for var in CONTEXT_VARS:
            self._context_vars_tokens[var] = var.set(var.get())

    def _reset_context_vars(self):
        # HACK: See _set_context_vars.
        for var, token in self._context_vars_tokens.items():
            var.reset(token)

        del self._context_vars_tokens[var]

    # WithInstrumentCallbacks requirement
    def on_new_record(
        self, func
    ) -> Iterable[core_instruments._RecordingContext]:
        """Called at the start of record creation.

        See
        [WithInstrumentCallbacks.on_new_record][trulens.core.instruments.WithInstrumentCallbacks.on_new_record].
        """
        ctx = self.recording_contexts.get(contextvars.Token.MISSING)

        while ctx is not contextvars.Token.MISSING:
            yield ctx
            ctx = ctx.token.old_value

    # WithInstrumentCallbacks requirement
    def on_add_record(
        self,
        ctx: core_instruments._RecordingContext,
        func: Callable,
        sig: Signature,
        bindings: BoundArguments,
        ret: Any,
        error: Any,
        perf: base_schema.Perf,
        cost: base_schema.Cost,
        existing_record: Optional[record_schema.Record] = None,
        final: bool = False,
    ) -> record_schema.Record:
        """Called by instrumented methods if they use _new_record to construct a
        "record call list.

        See
        [WithInstrumentCallbacks.on_add_record][trulens.core.instruments.WithInstrumentCallbacks.on_add_record].
        """

        def build_record(
            calls: Iterable[record_schema.RecordAppCall],
            record_metadata: serial_utils.JSON,
            existing_record: Optional[record_schema.Record] = None,
        ) -> record_schema.Record:
            calls = list(calls)

            assert len(calls) > 0, "No information recorded in call."

            if bindings is not None:
                if existing_record is None:
                    main_in = json_utils.jsonify(
                        self.main_input(func, sig, bindings)
                    )
                else:
                    main_in = existing_record.main_input
            else:
                main_in = None

            if error is None:
                assert bindings is not None, "No bindings despite no error."
                if final:
                    main_out = self.main_output(func, sig, bindings, ret)
                else:
                    main_out = f"TruLens: Record not yet finalized: {ret}"
            else:
                main_out = None

            updates = dict(
                main_input=main_in,
                main_output=json_utils.jsonify(main_out),
                main_error=json_utils.jsonify(error),
                calls=calls,
                cost=cost,
                perf=perf,
                app_id=self.app_id,
                tags=self.tags,
                meta=json_utils.jsonify(record_metadata),
            )

            if existing_record is not None:
                existing_record.update(**updates)
            else:
                existing_record = record_schema.Record(**updates)

            return existing_record

        # Finishing record needs to be done in a thread lock, done there:
        record = ctx.finish_record(
            build_record, existing_record=existing_record
        )

        if error is not None:
            # May block on DB.
            self._handle_error(record=record, error=error)
            raise error

        # Only continue with the feedback steps if the record is final.
        if not final:
            return record

        # Will block on DB, but not on feedback evaluation, depending on
        # FeedbackMode:
        record.feedback_and_future_results = self._handle_record(record=record)
        if record.feedback_and_future_results is not None:
            record.feedback_results = [
                tup[1] for tup in record.feedback_and_future_results
            ]

        if record.feedback_and_future_results is None:
            return record

        if self.feedback_mode == feedback_schema.FeedbackMode.WITH_APP_THREAD:
            # Add the record to ones with pending feedback.

            self.records_with_pending_feedback_results.add(record)

        elif self.feedback_mode == feedback_schema.FeedbackMode.WITH_APP:
            # If in blocking mode ("WITH_APP"), wait for feedbacks to finished
            # evaluating before returning the record.

            record.wait_for_feedback_results()

        return record

    def _check_instrumented(self, func):
        """
        Issue a warning and some instructions if a function that has not been
        instrumented is being used in a `with_` call.
        """

        if not isinstance(func, Callable):
            raise TypeError(
                f"Expected `func` to be a callable, but got {python_utils.class_name(type(func))}."
            )

        # If func is actually an object that implements __call__, check __call__
        # instead.
        if not (inspect.isfunction(func) or inspect.ismethod(func)):
            func = func.__call__

        if not python_utils.safe_hasattr(
            func, core_instruments.Instrument.INSTRUMENT
        ):
            if core_instruments.Instrument.INSTRUMENT in dir(func):
                # HACK009: Need to figure out the __call__ accesses by class
                # name/object name with relation to this check for
                # instrumentation because we keep hitting spurious warnings
                # here. This is a temporary workaround.
                return

            logger.warning(
                """
Function %s has not been instrumented. This may be ok if it will call a function
that has been instrumented exactly once. Otherwise unexpected results may
follow. You can use `AddInstruments.method` of `trulens.core.instruments` before
you use the `%s` wrapper to make sure `%s` does get instrumented. `%s` method
`print_instrumented` may be used to see methods that have been instrumented.
""",
                func,
                python_utils.class_name(self),
                python_utils.callable_name(func),
                python_utils.class_name(self),
            )

    async def awith_(
        self, func: asynchro_utils.CallableMaybeAwaitable[A, T], *args, **kwargs
    ) -> T:
        """Call the given async `func` with the given `*args` and `**kwargs`
        while recording, producing `func` results.

        The record of the computation is available through other means like the
        database or dashboard. If you need a record of this execution
        immediately, you can use `awith_record` or the `App` as a context
        manager instead.
        """

        res, _ = await self.awith_record(func, *args, **kwargs)

        return res

    async def with_(self, func: Callable[[A], T], *args, **kwargs) -> T:
        """Call the given async `func` with the given `*args` and `**kwargs`
        while recording, producing `func` results.

        The record of the computation is available through other means like the
        database or dashboard. If you need a record of this execution
        immediately, you can use `awith_record` or the `App` as a context
        manager instead.
        """

        res, _ = self.with_record(func, *args, **kwargs)

        return res

    def with_record(
        self,
        func: Callable[[A], T],
        *args,
        record_metadata: serial_utils.JSON = None,
        **kwargs,
    ) -> Tuple[T, record_schema.Record]:
        """
        Call the given `func` with the given `*args` and `**kwargs`, producing
        its results as well as a record of the execution.
        """

        if not isinstance(func, Callable):
            if hasattr(func, "__call__"):
                func = func.__call__

        self._check_instrumented(func)

        with self as ctx:
            ctx.record_metadata = record_metadata
            ret = func(*args, **kwargs)

        assert len(ctx.records) > 0, (
            f"Did not create any records. "
            f"This means that no instrumented methods were invoked in the process of calling {func}."
        )

        return ret, ctx.get()

    async def awith_record(
        self,
        func: Callable[[A], Awaitable[T]],
        *args,
        record_metadata: serial_utils.JSON = None,
        **kwargs,
    ) -> Tuple[T, record_schema.Record]:
        """
        Call the given `func` with the given `*args` and `**kwargs`, producing
        its results as well as a record of the execution.
        """

        if not isinstance(func, Callable):
            if hasattr(func, "__call__"):
                func = func.__call__

        self._check_instrumented(func)

        async with self as ctx:
            ctx.record_metadata = record_metadata
            ret = await func(*args, **kwargs)

        assert len(ctx.records) > 0, (
            f"Did not create any records. "
            f"This means that no instrumented methods were invoked in the process of calling {func}."
        )

        return ret, ctx.get()

    def _throw_dep_message(
        self, method, is_async: bool = False, with_record: bool = False
    ):
        # Raises a deprecation message for the various methods that pass through to
        # wrapped app while recording.

        cname = self.__class__.__name__

        iscall = method == "__call__"

        old_method = f"""{method}{"_with_record" if with_record else ""}"""
        if iscall:
            old_method = f"""call{"_with_record" if with_record else ""}"""
        new_method = f"""{"a" if is_async else ""}with_{"record" if with_record else ""}"""

        app_callable = f"""app.{method}"""
        if iscall:
            app_callable = "app"

        raise AttributeError(
            f"""
`{old_method}` is deprecated; To record results of your app's execution, use one of these options to invoke your app:
    (1) Use the `{"a" if is_async else ""}with_{"record" if with_record else ""}` method:
        ```python
        app # your app
        tru_app_recorder: {cname} = {cname}(app, ...)
        result{", record" if with_record else ""} = {"await " if is_async else ""}tru_app_recorder.{new_method}({app_callable}, ...args/kwargs-to-{app_callable}...)
        ```
    (2) Use {cname} as a context manager:
        ```python
        app # your app
        tru_app_recorder: {cname} = {cname}(app, ...)
        with tru_app_recorder{" as records" if with_record else ""}:
            result = {"await " if is_async else ""}{app_callable}(...args/kwargs-to-{app_callable}...)
        {"record = records.get()" if with_record else ""}
        ```
"""
        )

    def _add_future_feedback(
        self,
        future_or_result: Union[
            feedback_schema.FeedbackResult,
            Future[feedback_schema.FeedbackResult],
        ],
    ) -> None:
        """
        Callback used to add feedback results to the database once they are
        done.

        See [_handle_record][trulens.core.app.App._handle_record].
        """

        if isinstance(future_or_result, Future):
            res = future_or_result.result()
        else:
            res = future_or_result

        self.connector.add_feedback(res)

    def _handle_record(
        self,
        record: record_schema.Record,
        feedback_mode: Optional[feedback_schema.FeedbackMode] = None,
    ) -> Optional[
        List[
            Tuple[
                core_feedback.Feedback,
                Future[feedback_schema.FeedbackResult],
            ]
        ]
    ]:
        """
        Write out record-related info to database if set and schedule feedback
        functions to be evaluated. If feedback_mode is provided, will use that
        mode instead of the one provided to constructor.
        """

        if feedback_mode is None:
            feedback_mode = self.feedback_mode

        if self.feedback_mode is None:
            return None

        # If in buffered mode, call add record nowait.
        if self.record_ingest_mode == app_schema.RecordIngestMode.BUFFERED:
            self.connector.add_record_nowait(record=record)
            return

        # Need to add record to db before evaluating feedback functions.
        record_id = self.connector.add_record(record=record)

        if len(self.feedbacks) == 0:
            return []

        if feedback_mode == feedback_schema.FeedbackMode.NONE:
            # Do not run any feedbacks in this case (now or deferred).
            return None

        if feedback_mode == feedback_schema.FeedbackMode.DEFERRED:
            # Run all feedbacks as deferred.
            deferred_feedbacks = self.feedbacks
            undeferred_feedbacks = []
        else:
            # Run only the feedbacks to be run in Snowflake as deferred.
            deferred_feedbacks = []
            undeferred_feedbacks = []
            for f in self.feedbacks:
                if (
                    f.run_location
                    == feedback_schema.FeedbackRunLocation.SNOWFLAKE
                ):
                    deferred_feedbacks.append(f)
                else:
                    undeferred_feedbacks.append(f)

        # Insert into the feedback table the deferred feedbacks.
        for f in deferred_feedbacks:
            self.connector.db.insert_feedback(
                feedback_schema.FeedbackResult(
                    name=f.name,
                    record_id=record_id,
                    feedback_definition_id=f.feedback_definition_id,
                )
            )
        # Compute the undeferred feedbacks.
        return self._submit_feedback_functions(
            record=record,
            feedback_functions=undeferred_feedbacks,
            app=self,
            connector=self.connector,
            on_done=self._add_future_feedback,
        )

    def _handle_error(self, record: record_schema.Record, error: Exception):
        if self.connector is None:
            return

    def __getattr__(self, __name: str) -> Any:
        # A message for cases where a user calls something that the wrapped app
        # contains. We do not support this form of pass-through calls anymore.

        try:
            # Some odd interaction with pydantic.PrivateAttr causes this handler
            # to be called for private attributes even though they exist. So we
            # double check here with pydantic's getattr.
            return pydantic.BaseModel.__getattr__(self, __name)
        except AttributeError:
            pass

        app = self.app

        if python_utils.safe_hasattr(app, __name):
            msg = ATTRIBUTE_ERROR_MESSAGE.format(
                attribute_name=__name,
                class_name=type(self).__name__,
                app_class_name=type(self.app).__name__,
            )
            raise AttributeError(msg)

        else:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{__name}'"
            )

    def dummy_record(
        self,
        cost: base_schema.Cost = base_schema.Cost(),
        perf: base_schema.Perf = base_schema.Perf.now(),
        ts: datetime.datetime = datetime.datetime.now(),
        main_input: str = "main_input are strings.",
        main_output: str = "main_output are strings.",
        main_error: str = "main_error are strings.",
        meta: Dict = {"metakey": "meta are dicts"},
        tags: str = "tags are strings",
    ) -> record_schema.Record:
        """Create a dummy record with some of the expected structure without
        actually invoking the app.

        The record is a guess of what an actual record might look like but will
        be missing information that can only be determined after a call is made.

        All args are [Record][trulens.core.schema.record.Record] fields except these:

            - `record_id` is generated using the default id naming schema.
            - `app_id` is taken from this recorder.
            - `calls` field is constructed based on instrumented methods.
        """

        calls = []

        for methods in self.instrumented_methods.values():
            for func, lens in methods.items():
                component = lens.get_sole_item(self)

                if not hasattr(component, func.__name__):
                    continue
                method = getattr(component, func.__name__)

                sig = inspect.signature(method)

                method_serial = pyschema_utils.FunctionOrMethod.of_callable(
                    method
                )

                sample_args = {}
                for p in sig.parameters.values():
                    if p.default == inspect.Parameter.empty:
                        sample_args[p.name] = None
                    else:
                        sample_args[p.name] = p.default

                sample_call = record_schema.RecordAppCall(
                    stack=[
                        record_schema.RecordAppCallMethod(
                            path=lens, method=method_serial
                        )
                    ],
                    args=sample_args,
                    rets=None,
                    pid=0,
                    tid=0,
                )

                calls.append(sample_call)

        return record_schema.Record(
            app_id=self.app_id,
            calls=calls,
            cost=cost,
            perf=perf,
            ts=ts,
            main_input=main_input,
            main_output=main_output,
            main_error=main_error,
            meta=meta,
            tags=tags,
        )

    def instrumented(self) -> Iterable[Tuple[serial_utils.Lens, ComponentView]]:
        """
        Iteration over instrumented components and their categories.
        """

        for q, c in instrumented_component_views(self.model_dump()):
            # Add the chain indicator so the resulting paths can be specified
            # for feedback selectors.
            q = serial_utils.Lens(
                path=(
                    serial_utils.GetItemOrAttribute(
                        item_or_attribute="__app__"
                    ),
                )
                + q.path
            )
            yield q, c

    def print_instrumented(self) -> None:
        """Print the instrumented components and methods."""

        print("Components:")
        self.print_instrumented_components()
        print("\nMethods:")
        self.print_instrumented_methods()

    def format_instrumented_methods(self) -> str:
        """Build a string containing a listing of instrumented methods."""

        return "\n".join(
            f"Object at 0x{obj:x}:\n\t"
            + "\n\t".join(
                f"{m} with path {select_schema.Select.App + path}"
                for m, path in p.items()
            )
            for obj, p in self.instrumented_methods.items()
        )

    def print_instrumented_methods(self) -> None:
        """Print instrumented methods."""

        logger.info(self.format_instrumented_methods())

    def print_instrumented_components(self) -> None:
        """Print instrumented components and their categories."""

        object_strings = []

        for t in self.instrumented():
            path = serial_utils.Lens(t[0].path[1:])
            obj = next(iter(path.get(self)))
            object_strings.append(
                f"\t{type(obj).__name__} ({t[1].__class__.__name__}) at 0x{id(obj):x} with path {str(t[0])}"
            )

        logger.info("\n".join(object_strings))

    def _check_snowflake_dao(self):
        if (
            not hasattr(self, "snowflake_app_dao")
            or self.snowflake_app_dao is None
        ):
            msg = (
                "This API requires a Snowpark session to initialize snowflake-specific DAO instance. Please initialize App with "
                "object_type='EXTERNAL AGENT' and a valid snowpark_session."
            )
            logger.error(msg)
            raise ValueError(msg)

    def add_run(self, run_config: RunConfig) -> Union[Run, None]:
        """add a new run to the snowflake App (if not already exists)

        Args:
            run_config (RunConfig):  Run config
            input_df (Optional[pd.DataFrame]): optional input dataset

        Returns:
            Run: Run instance
        """

        self._check_snowflake_dao()

        try:
            run_metadata_df = self.snowflake_run_dao.create_new_run(
                object_name=self.snowflake_object_name,
                object_type=self.snowflake_object_type,
                object_version=self.snowflake_object_version,
                run_name=run_config.run_name,
                dataset_name=run_config.dataset_name,
                source_type=run_config.source_type,
                dataset_spec=validate_dataset_spec(run_config.dataset_spec),
                description=run_config.description,
                label=run_config.label,
                llm_judge_name=run_config.llm_judge_name,
            )

            return Run.from_metadata_df(
                run_metadata_df,
                {
                    "app": self,
                    "main_method_name": self.main_method_name,
                    "run_dao": self.snowflake_run_dao,
                    "tru_session": self.session,
                },
            )

        except Exception as e:
            if "already exists." in str(e):
                logger.debug(f"Run {run_config.run_name} already exists.")
                return self.get_run(run_config.run_name)
            else:
                logger.error(f"Failed to add run {run_config.run_name}. {e}")
                raise

    def get_run(self, run_name: str) -> Run:
        """Retrieve a run by name.

        Args:
            run_name (str): unique name of the run

        Returns:
            Run: Run instance
        """
        self._check_snowflake_dao()
        run_metadata_df = self.snowflake_run_dao.get_run(
            run_name=run_name,
            object_name=self.snowflake_object_name,
            object_type=self.snowflake_object_type,
            object_version=self.snowflake_object_version,
        )
        if run_metadata_df.empty:
            raise ValueError(f"Run {run_name} not found.")

        return Run.from_metadata_df(
            run_metadata_df,
            {
                "app": self,
                "main_method_name": self.main_method_name,
                "run_dao": self.snowflake_run_dao,
                "tru_session": self.session,
            },
        )

    def list_runs(self) -> List[Run]:
        """Retrieve all runs belong to the snowflake App.

        Returns:
            List[Run]: List of Run instances
        """
        self._check_snowflake_dao()
        run_metadata_df = self.snowflake_run_dao.list_all_runs(
            object_name=self.snowflake_object_name,
            object_type=self.snowflake_object_type,
        )
        runs = []

        if run_metadata_df.empty:
            return runs

        all_runs_lst = json.loads(run_metadata_df.iloc[0].iloc[-1])

        # return all_runs_lst
        for run_dict in all_runs_lst:
            runs.append(
                Run.from_metadata_df(
                    pd.DataFrame({"col": [json.dumps(run_dict)]}),
                    {
                        "app": self.app,
                        "main_method_name": self.main_method_name,
                        "run_dao": self.snowflake_run_dao,
                        "tru_session": self.session,
                        "object_name": self.snowflake_object_name,
                        "object_type": self.snowflake_object_type,
                    },
                )
            )
        return runs

    def delete(self) -> None:
        """Delete the snowflake App (external agent) in snowflake. All versions will be removed"""
        self._check_snowflake_dao()
        self.snowflake_app_dao.drop_agent(self.snowflake_object_name)

    def delete_version(self) -> None:
        """Delete the current version of the snowflake App (external agent) in snowflake.
        Only the non-default version can be deleted.
        """
        self._check_snowflake_dao()
        self.snowflake_app_dao.drop_current_version(self.snowflake_object_name)

    def run(self, run_name: str):
        if self.session.experimental_feature(
            core_experimental.Feature.OTEL_TRACING
        ):
            from trulens.core.otel.instrument import OtelRecordingContext

            return OtelRecordingContext(
                app_name=self.app_name,
                app_version=self.app_version,
                run_name=run_name,
                input_id=None,
            )
        raise NotImplementedError(
            "This feature is not yet implemented for non-OTEL TruLens!"
        )

    def input(self, input_id: str):
        if self.session.experimental_feature(
            core_experimental.Feature.OTEL_TRACING
        ):
            from trulens.core.otel.instrument import OtelRecordingContext

            return OtelRecordingContext(
                app_name=self.app_name,
                app_version=self.app_version,
                run_name=None,
                input_id=input_id,
            )
        raise NotImplementedError(
            "This feature is not yet implemented for non-OTEL TruLens!"
        )

    def instrumented_invoke_main_method(
        self,
        run_name: str,
        input_id: str,
        ground_truth_output: Optional[str] = None,
        main_method_args: Optional[Sequence[Any]] = None,
        main_method_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Any:
        if self.session.experimental_feature(
            core_experimental.Feature.OTEL_TRACING
        ):
            from trulens.core.otel.instrument import OtelRecordingContext

            with OtelRecordingContext(
                app_name=self.app_name,
                app_version=self.app_version,
                run_name=run_name,
                input_id=input_id,
                ground_truth_output=ground_truth_output,
            ):
                f = getattr(self.app, self.main_method_name)
                if main_method_args is None:
                    main_method_args = ()
                if main_method_kwargs is None:
                    main_method_kwargs = {}
                return f(*main_method_args, **main_method_kwargs)
        raise NotImplementedError(
            "This feature is not yet implemented for non-OTEL TruLens!"
        )


# NOTE: Cannot App.model_rebuild here due to circular imports involving mod_session.TruSession
# and database.base.DB. Will rebuild each App subclass instead.



from __future__ import annotations

import abc
import functools
from sqlite3 import Connection as SQLite3Connection
from typing import ClassVar, Dict, Generic, Type, TypeVar

from sqlalchemy import JSON
from sqlalchemy import TIMESTAMP
from sqlalchemy import VARCHAR
from sqlalchemy import Column
from sqlalchemy import Engine
from sqlalchemy import Enum
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Text
from sqlalchemy import UniqueConstraint
from sqlalchemy import event
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import backref
from sqlalchemy.orm import configure_mappers
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.schema import MetaData
from trulens.core.database import base as core_db
from trulens.core.schema import app as app_schema
from trulens.core.schema import dataset as dataset_schema
from trulens.core.schema import event as event_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import groundtruth as groundtruth_schema
from trulens.core.schema import record as record_schema
from trulens.core.utils import json as json_utils

TYPE_JSON = Text
"""Database type for JSON fields."""

TYPE_TIMESTAMP = Float
"""Database type for timestamps."""

TYPE_ENUM = Text
"""Database type for enum fields."""

TYPE_ID = VARCHAR(256)
"""Database type for unique IDs."""


class BaseWithTablePrefix:  # to be mixed into DeclarativeBase or new_declarative_base()
    # Only for type hints or isinstance, issubclass checks.
    """ORM base class except with `__tablename__` defined in terms
    of a base name and a prefix.

    A subclass should set _table_base_name and/or _table_prefix. If it does not
    set both, make sure to set `__abstract__ = True`. Current design has
    subclasses set `_table_base_name` and then subclasses of that subclass
    setting `_table_prefix` as in `make_orm_for_prefix`.
    """

    # https://stackoverflow.com/questions/38245145/how-to-set-common-prefix-for-all-tables-in-sqlalchemy
    # Needed for sqlaclhemy to prevent it from creating a table for this class
    # before the two following attributes are set which we do in subclasses later.
    __abstract__ = True

    _table_base_name: str = "not set"
    """Base name for the table.

    Will be prefixed by the prefix to create table names. This should be set by
    subclasses.
    """

    _table_prefix: str = ""
    """Prefix for the table name.

    This should be set by subclasses of subclasses of this class.
    """

    @declared_attr.directive
    def __tablename__(cls) -> str:
        return cls._table_prefix + cls._table_base_name


T = TypeVar("T", bound=BaseWithTablePrefix)


# NOTE: lru_cache is important here as we don't want to create multiple classes
# of the same name for the same table name prefix as sqlalchemy will complain
# one some of our migration tools will not work.
@functools.lru_cache
def new_base(prefix: str) -> Type[T]:
    """Create a new base class for ORM classes.

    Note: This is a function to be able to define classes extending different
    SQLAlchemy declarative bases. Each different such bases has a different set
    of mappings from classes to table names. If we only had one of these, our
    code will never be able to have two different sets of mappings at the same
    time. We need to be able to have multiple mappings for performing things
    such as database migrations and database copying from one database
    configuration to another.
    """

    base = declarative_base()
    return type(
        f"BaseWithTablePrefix{prefix}",
        (base, BaseWithTablePrefix),
        {
            "_table_prefix": prefix,
            "__abstract__": True,  # stay abstract until _table_base_name is set in a subclass
        },
    )


class ORM(abc.ABC, Generic[T]):
    """Abstract definition of a container for ORM classes."""

    registry: Dict[str, Type[T]]
    metadata: MetaData

    AppDefinition: Type[T]
    FeedbackDefinition: Type[T]
    Record: Type[T]
    FeedbackResult: Type[T]
    GroundTruth: Type[T]
    Dataset: Type[T]
    Event: Type[T]


def new_orm(base: Type[T], prefix: str = "trulens_") -> Type[ORM[T]]:
    """Create a new orm container from the given base table class."""

    class NewORM(ORM):
        """Container for ORM classes.

        Needs to be extended with classes that set table prefix.

        Warning:
            The relationships between tables established in the classes in this
            container refer to class names i.e. "AppDefinition" hence these are
            important and need to stay consistent between definition of one and
            relationships in another.
        """

        registry: Dict[str, base] = base.registry._class_registry
        """Table name to ORM class mapping for tables used by trulens.

        This can be used to iterate through all classes/tables.
        """

        metadata: MetaData = base.metadata
        """SqlAlchemy metadata object for tables used by trulens."""

        class AppDefinition(base):
            """ORM class for [AppDefinition][trulens.core.schema.app.AppDefinition].

            Warning:
                We don't use any of the typical ORM features and this class is only
                used as a schema to interact with database through SQLAlchemy.
            """

            _table_base_name: ClassVar[str] = "apps"

            app_id = Column(
                VARCHAR(256),
                nullable=False,
                primary_key=True,
            )
            app_name = Column(
                VARCHAR(256), nullable=False, default="default_app"
            )
            app_version = Column(VARCHAR(256), nullable=False, default="base")
            app_json = Column(TYPE_JSON, nullable=False)
            __table_args__ = (UniqueConstraint("app_name", "app_version"),)

            # records via one-to-many on Record.app_id
            # feedback_results via one-to-many on FeedbackResult.record_id

            @classmethod
            def parse(
                cls,
                obj: app_schema.AppDefinition,
                redact_keys: bool = False,
            ) -> ORM.AppDefinition:
                return cls(
                    app_id=obj.app_id,
                    app_name=obj.app_name,
                    app_version=obj.app_version,
                    app_json=obj.model_dump_json(redact_keys=redact_keys),
                )

        class FeedbackDefinition(base):
            """ORM class for
            [FeedbackDefinition][trulens.core.schema.feedback.FeedbackDefinition]."""

            _table_base_name = "feedback_defs"

            feedback_definition_id = Column(
                TYPE_ID, nullable=False, primary_key=True
            )
            run_location = Column(TYPE_ENUM, nullable=True)
            feedback_json = Column(TYPE_JSON, nullable=False)

            # feedback_results via one-to-many on FeedbackResult.feedback_definition_id

            @classmethod
            def parse(
                cls,
                obj: feedback_schema.FeedbackDefinition,
                redact_keys: bool = False,
            ) -> ORM.FeedbackDefinition:
                return cls(
                    feedback_definition_id=obj.feedback_definition_id,
                    run_location=None
                    if obj.run_location is None
                    else obj.run_location.value,
                    feedback_json=json_utils.json_str_of_obj(
                        obj, redact_keys=redact_keys
                    ),
                )

        class Record(base):
            """ORM class for [Record][trulens.core.schema.record.Record]."""

            _table_base_name = "records"

            record_id = Column(TYPE_ID, nullable=False, primary_key=True)

            # Couldn't figure out how to avoid using prefix in ForeignKey here
            # and in the other classes below.
            app_id = Column(
                TYPE_ID, ForeignKey(f"{prefix}apps.app_id"), nullable=False
            )

            input = Column(Text)
            output = Column(Text)
            record_json = Column(TYPE_JSON, nullable=False)
            tags = Column(Text, nullable=False)
            ts = Column(TYPE_TIMESTAMP, nullable=False)
            cost_json = Column(TYPE_JSON, nullable=False)
            perf_json = Column(TYPE_JSON, nullable=False)

            app = relationship(
                "AppDefinition",
                backref=backref(
                    "records", cascade="all,delete", order_by=(ts, record_id)
                ),
            )
            # NOTE(backref order_by): The order_by must be inside the backref as
            # it refers to the ordering of AppDefinition.records, not of
            # Record.app. Doing the opposite can produce SQL errors later on but
            # no warning will be given at schema creation time.

            @classmethod
            def parse(
                cls, obj: record_schema.Record, redact_keys: bool = False
            ) -> ORM.Record:
                return cls(
                    record_id=obj.record_id,
                    app_id=obj.app_id,
                    input=json_utils.json_str_of_obj(
                        obj.main_input, redact_keys=redact_keys
                    ),
                    output=json_utils.json_str_of_obj(
                        obj.main_output, redact_keys=redact_keys
                    ),
                    record_json=json_utils.json_str_of_obj(
                        obj, redact_keys=redact_keys
                    ),
                    tags=obj.tags,
                    ts=obj.ts.timestamp(),
                    cost_json=json_utils.json_str_of_obj(
                        obj.cost, redact_keys=redact_keys
                    ),
                    perf_json=json_utils.json_str_of_obj(
                        obj.perf, redact_keys=redact_keys
                    ),
                )

        class FeedbackResult(base):
            """ORM class for
            [FeedbackResult][trulens.core.schema.feedback.FeedbackResult]."""

            _table_base_name = "feedbacks"

            feedback_result_id = Column(
                TYPE_ID, nullable=False, primary_key=True
            )
            record_id = Column(
                TYPE_ID,
                ForeignKey(f"{prefix}records.record_id"),
                nullable=False,
            )
            feedback_definition_id = Column(
                TYPE_ID,
                ForeignKey(f"{prefix}feedback_defs.feedback_definition_id"),
                nullable=False,
            )
            last_ts = Column(TYPE_TIMESTAMP, nullable=False)
            status = Column(TYPE_ENUM, nullable=False)
            error = Column(Text)
            calls_json = Column(TYPE_JSON, nullable=False)
            result = Column(Float)
            name = Column(Text, nullable=False)
            cost_json = Column(TYPE_JSON, nullable=False)
            multi_result = Column(TYPE_JSON)

            record = relationship(
                "Record",
                backref=backref(
                    "feedback_results",
                    cascade="all,delete",
                    order_by=(last_ts, feedback_result_id),
                ),
            )
            # See NOTE(backref order_by).

            feedback_definition = relationship(
                "FeedbackDefinition",
                backref=backref(
                    "feedback_results",
                    cascade="all,delete",
                    order_by=(last_ts, feedback_result_id),
                ),
            )
            # See NOTE(backref order_by).

            @classmethod
            def parse(
                cls,
                obj: feedback_schema.FeedbackResult,
                redact_keys: bool = False,
            ) -> ORM.FeedbackResult:
                return cls(
                    feedback_result_id=obj.feedback_result_id,
                    record_id=obj.record_id,
                    feedback_definition_id=obj.feedback_definition_id,
                    last_ts=obj.last_ts.timestamp(),
                    status=obj.status.value,
                    error=obj.error,
                    calls_json=json_utils.json_str_of_obj(
                        dict(calls=obj.calls), redact_keys=redact_keys
                    ),
                    result=obj.result,
                    name=obj.name,
                    cost_json=json_utils.json_str_of_obj(
                        obj.cost, redact_keys=redact_keys
                    ),
                    multi_result=obj.multi_result,
                )

        class GroundTruth(base):
            """ORM class for
            [GroundTruth][trulens.core.schema.groundtruth.GroundTruth]."""

            _table_base_name = "ground_truth"

            ground_truth_id = Column(TYPE_ID, nullable=False, primary_key=True)
            dataset_id = Column(
                Text, ForeignKey(f"{prefix}dataset.dataset_id"), nullable=False
            )
            ground_truth_json = Column(TYPE_JSON, nullable=False)

            dataset = relationship(
                "Dataset",
                backref=backref(
                    "ground_truths",
                    cascade="all,delete",
                    order_by=ground_truth_id,
                ),
            )
            # See NOTE(backref order_by).

            @classmethod
            def parse(
                cls,
                obj: groundtruth_schema.GroundTruth,
                redact_keys: bool = False,
            ) -> ORM.GroundTruth:
                return cls(
                    ground_truth_id=obj.ground_truth_id,
                    ground_truth_json=obj.model_dump_json(
                        redact_keys=redact_keys
                    ),
                    dataset_id=obj.dataset_id,
                )

        class Dataset(base):
            """
            ORM class for [Dataset][trulens.core.schema.dataset.Dataset].

            Warning:
                We don't use any of the typical ORM features and this class is only
                used as a schema to interact with database through SQLAlchemy.
            """

            _table_base_name = "dataset"

            dataset_id = Column(TYPE_ID, nullable=False, primary_key=True)
            dataset_json = Column(TYPE_JSON, nullable=False)

            @classmethod
            def parse(
                cls,
                obj: dataset_schema.Dataset,
                redact_keys: bool = False,
            ) -> ORM.Dataset:
                return cls(
                    dataset_id=obj.dataset_id,
                    dataset_json=obj.model_dump_json(redact_keys=redact_keys),
                )

        class Event(base):
            """
            ORM class for OTEL traces/spans. This should be kept in line with the event table
            https://docs.snowflake.com/en/developer-guide/logging-tracing/event-table-columns#data-for-trace-events
            https://docs.snowflake.com/en/developer-guide/logging-tracing/event-table-columns#label-event-table-record-column-span
            """

            _table_base_name = "events"

            record = Column(JSON, nullable=False)
            """
            For a span, this is an object that includes:
            - name: the function/procedure that emitted the data
            - kind: SPAN_KIND_TRULENS
            - parent_span_id: the unique identifier for the parent span
            - status: STATUS_CODE_ERROR when the span corresponds to an unhandled exception. Otherwise, STATUS_CODE_UNSET.
            """

            event_id = Column(TYPE_ID, nullable=False, primary_key=True)
            """
            Used as primary key for the schema. Not technically present in the event table schema,
            but having it here makes it easier to work with the ORM.
            """

            record_attributes = Column(JSON, nullable=False)
            """
            Attributes of the record that can either come from the user, or based on the TruLens semantic conventions.
            """

            record_type = Column(
                Enum(event_schema.EventRecordType), nullable=False
            )
            """
            Specifies the kind of record specified by this row. This will always be "SPAN" for TruLens.
            """

            resource_attributes = Column(JSON, nullable=False)
            """
            Reserved.
            """

            start_timestamp = Column(TIMESTAMP, nullable=False)
            """
            The timestamp when the span started. This is a UNIX timestamp in milliseconds.
            Note: The Snowflake event table uses the TIMESTAMP_NTZ data type for this column.
            """

            timestamp = Column(TIMESTAMP, nullable=False)
            """
            The timestamp when the span concluded. This is a UNIX timestamp in milliseconds.
            Note: The Snowflake event table uses the TIMESTAMP_NTZ data type for this column.
            """

            trace = Column(JSON, nullable=False)
            """
            Contains the span context, including the trace_id, parent_id, and span_id for the span.
            """

            @classmethod
            def parse(
                cls,
                obj: event_schema.Event,
                redact_keys: bool = False,
            ) -> ORM.EventTable:
                return cls(
                    event_id=obj.event_id,
                    record=obj.record,
                    record_attributes=obj.record_attributes,
                    record_type=obj.record_type,
                    resource_attributes=obj.resource_attributes,
                    start_timestamp=obj.start_timestamp,
                    timestamp=obj.timestamp,
                    trace=obj.trace,
                )

    configure_mappers()  # IMPORTANT
    # Without the above, orm class attributes which are defined using backref
    # will not be visible, i.e. orm.AppDefinition.records.

    # base.registry.configure()

    return NewORM


# NOTE: lru_cache is important here as we don't want to create multiple classes for
# the same table name as sqlalchemy will complain.
@functools.lru_cache
def make_base_for_prefix(
    base: Type[T],
    table_prefix: str = core_db.DEFAULT_DATABASE_PREFIX,
) -> Type[T]:
    """
    Create a base class for ORM classes with the given table name prefix.

    Args:
        base: Base class to extend. Should be a subclass of
            [BaseWithTablePrefix][trulens.core.database.orm.BaseWithTablePrefix].

        table_prefix: Prefix to use for table names.

    Returns:
        A class that extends `base_type` and sets the table prefix to `table_prefix`.
    """

    if not hasattr(base, "_table_base_name"):
        raise ValueError(
            "Expected `base` to be a subclass of `BaseWithTablePrefix`."
        )

    # sqlalchemy stores a mapping of class names to the classes we defined in
    # the ORM above. Here we want to create a class with the specific name
    # matching base_type hence use `type` instead of `class SomeName: ...`.
    return type(base.__name__, (base,), {"_table_prefix": table_prefix})


# NOTE: lru_cache is important here as we don't want to create multiple classes for
# the same table name as sqlalchemy will complain.
@functools.lru_cache
def make_orm_for_prefix(
    table_prefix: str = core_db.DEFAULT_DATABASE_PREFIX,
) -> Type[ORM[T]]:
    """
    Make a container for ORM classes.

    This is done so that we can use a dynamic table name prefix and make the ORM
    classes based on that.

    Args:
        table_prefix: Prefix to use for table names.
    """

    base: Type[T] = new_base(prefix=table_prefix)

    return new_orm(base, prefix=table_prefix)


@event.listens_for(Engine, "connect")
def _set_sqlite_pragma(dbapi_connection, _):
    if isinstance(dbapi_connection, SQLite3Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON;")
        cursor.close()



"""Serializable record-related classes."""

from __future__ import annotations

from concurrent.futures import as_completed
import datetime
import logging
from typing import (
    ClassVar,
    Dict,
    Hashable,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
)

from munch import Munch as Bunch
import pydantic
from trulens.core._utils.pycompat import Future  # import style exception
from trulens.core.schema import base as base_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import select as select_schema
from trulens.core.schema import types as types_schema
from trulens.core.utils import json as json_utils
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import threading as threading_utils

T = TypeVar("T")

logger = logging.getLogger(__name__)


class RecordAppCallMethod(serial_utils.SerialModel):
    """Method information for the stacks inside `RecordAppCall`."""

    def __str__(self):
        return f"{self.path}.{self.method.name}"

    path: serial_utils.Lens
    """Path to the method in the app's structure."""

    method: pyschema_utils.Method
    """The method that was called."""


class RecordAppCall(serial_utils.SerialModel):
    """Info regarding each instrumented method call."""

    def __str__(self):
        stack = " -> ".join(map(str, self.stack))
        return f"RecordAppCall: {stack}"

    call_id: types_schema.CallID = pydantic.Field(
        default_factory=types_schema.new_call_id
    )
    """Unique identifier for this call.

    This is shared across different instances of
    [RecordAppCall][trulens.core.schema.record.RecordAppCall] if they refer to
    the same python method call. This may happen if multiple recorders capture
    the call in which case they will each have a different
    [RecordAppCall][trulens.core.schema.record.RecordAppCall] but the
    [call_id][trulens.core.schema.record.RecordAppCall.call_id] will be the
    same.
    """

    stack: List[RecordAppCallMethod]
    """Call stack but only containing paths of instrumented apps/other objects."""

    args: serial_utils.JSON
    """Arguments to the instrumented method."""

    rets: Optional[serial_utils.JSON] = None
    """Returns of the instrumented method if successful.

    Sometimes this is a dict, sometimes a sequence, and sometimes a base value.
    """

    error: Optional[str] = None
    """Error message if call raised exception."""

    perf: Optional[base_schema.Perf] = None
    """Timestamps tracking entrance and exit of the instrumented method."""

    pid: int
    """Process id."""

    tid: int
    """Thread id."""

    @property
    def top(self) -> RecordAppCallMethod:
        """The top of the stack."""

        return self.stack[-1]

    @property
    def method(self) -> pyschema_utils.Method:
        """The method at the top of the stack."""

        return self.top.method


class Record(serial_utils.SerialModel, Hashable):
    """The record of a single main method call.

    Note:
        This class will be renamed to `Trace` in the future.
    """

    def __str__(self):
        ret = f"Record({self.record_id}) with {len(self.calls)} calls:\n"
        for call in self.calls:
            ret += "  " + str(call) + "\n"

        return ret

    model_config: ClassVar[dict] = {
        # for `Future[FeedbackResult]`
        "arbitrary_types_allowed": True
    }

    record_id: types_schema.RecordID
    """Unique identifier for this record."""

    app_id: types_schema.AppID
    """The app that produced this record."""

    cost: Optional[base_schema.Cost] = None
    """Costs associated with the record."""

    perf: Optional[base_schema.Perf] = None
    """Performance information."""

    ts: datetime.datetime = pydantic.Field(
        default_factory=datetime.datetime.now
    )
    """Timestamp of last update.

    This is usually set whenever a record is changed in any way."""

    tags: Optional[str] = ""
    """Tags for the record."""

    meta: Optional[serial_utils.JSON] = None
    """Metadata for the record."""

    main_input: Optional[serial_utils.JSON] = None
    """The app's main input."""

    main_output: Optional[serial_utils.JSON] = None  # if no error
    """The app's main output if there was no error."""

    main_error: Optional[serial_utils.JSON] = None  # if error
    """The app's main error if there was an error."""

    calls: List[RecordAppCall] = []
    """The collection of calls recorded.

    Note that these can be converted into a json structure with the same paths
    as the app that generated this record via `layout_calls_as_app`.

    Invariant: calls are ordered by `.perf.end_time`.
    """

    feedback_and_future_results: Optional[
        List[
            Tuple[
                feedback_schema.FeedbackDefinition,
                Future[feedback_schema.FeedbackResult],
            ]
        ]
    ] = pydantic.Field(None, exclude=True)
    """Map of feedbacks to the futures for of their results.

    These are only filled for records that were just produced. This will not
    be filled in when read from database. Also, will not fill in when using
    `FeedbackMode.DEFERRED`.
    """

    feedback_results: Optional[List[Future[feedback_schema.FeedbackResult]]] = (
        pydantic.Field(None, exclude=True)
    )
    """Only the futures part of the above for backwards compatibility."""

    @property
    def feedback_results_as_completed(
        self,
    ) -> Iterable[feedback_schema.FeedbackResult]:
        """Generate feedback results as they are completed.

        Wraps
        [feedback_results][trulens.core.schema.record.Record.feedback_results]
        in [as_completed][concurrent.futures.as_completed].
        """

        if self.feedback_results is None:
            return

        for f in as_completed(self.feedback_results):
            yield f.result()

    def __init__(
        self,
        record_id: Optional[types_schema.RecordID] = None,
        calls: Optional[List[RecordAppCall]] = None,
        **kwargs,
    ):
        super(serial_utils.SerialModel, self).__init__(
            record_id="temporary", calls=calls, **kwargs
        )

        # Resetting the calls in sorted order here. Note that we are calling
        # init with calls above to make sure they get converted to RecordAppCall
        # from dicts. We then sort those in the next step:
        if calls is not None:
            self.calls = sorted(
                self.calls,
                key=lambda call: call.perf.end_time
                if call.perf is not None
                else datetime.datetime.max,
            )

        if record_id is None:
            record_id = json_utils.obj_id_of_obj(
                json_utils.jsonify(self), prefix="record"
            )

        self.record_id = record_id

    def __hash__(self):
        return hash(self.record_id)

    def wait_for_feedback_results(
        self, feedback_timeout: Optional[float] = None
    ) -> Dict[
        feedback_schema.FeedbackDefinition,
        feedback_schema.FeedbackResult,
    ]:
        """Wait for feedback results to finish.

        Args:
            feedback_timeout: Timeout in seconds for each feedback function. If
                not given, will use the default timeout
                `trulens.core.utils.threading.TP.DEBUG_TIMEOUT`.

        Returns:
            A mapping of feedback functions to their results.
        """

        if feedback_timeout is None:
            feedback_timeout = threading_utils.TP.DEBUG_TIMEOUT

        if self.feedback_and_future_results is None:
            return {}

        ret = {}

        for feedback_def, future_result in self.feedback_and_future_results:
            try:
                feedback_result = future_result.result(timeout=feedback_timeout)
                ret[feedback_def] = feedback_result
            except TimeoutError:
                logger.error(
                    "Timeout waiting for feedback result for %s.",
                    feedback_def.name,
                )

        return ret

    def get(self, path: serial_utils.Lens) -> Optional[T]:
        """Get a value from the record using a path.

        Args:
            path: Path to the value.
        """

        layout = self.layout_calls_as_app()

        if len(path.path) == 0:
            return layout

        if path.path[0] == select_schema.Select.App.path[0]:
            raise ValueError("This path references an app, not a record.")
        elif path.path[0] == select_schema.Select.Record.path[0]:
            path = serial_utils.Lens(path.path[1:])

        return path.get_sole_item(obj=layout)

    def layout_calls_as_app(self) -> Bunch:
        """Layout the calls in this record into the structure that follows that of
        the app that created this record.

        This uses the paths stored in each
        [RecordAppCall][trulens.core.schema.record.RecordAppCall] which are paths into
        the app.

        Note: We cannot create a validated
        [AppDefinition][trulens.core.schema.app.AppDefinition] class (or subclass)
        object here as the layout of records differ in these ways:

        - Records do not include anything that is not an instrumented method
          hence have most of the structure of a app missing.

        - Records have RecordAppCall as their leafs where method definitions
          would be in the AppDefinition structure.
        """

        ret = Bunch(**self.model_dump())

        for call in self.calls:
            # Info about the method call is at the top of the stack
            frame_info = call.top

            # Adds another attribute to path, from method name:
            path = frame_info.path + serial_utils.GetItemOrAttribute(
                item_or_attribute=frame_info.method.name
            )

            if path.exists(obj=ret):
                existing = path.get_sole_item(obj=ret)
                ret = path.set(obj=ret, val=existing + [call])
            else:
                ret = path.set(obj=ret, val=[call])

        ret.spans = {}
        return ret


# HACK013: Need these if using __future__.annotations .
RecordAppCallMethod.model_rebuild()
Record.model_rebuild()
RecordAppCall.model_rebuild()



"""Custom class application

This wrapper is the most flexible option for instrumenting an application, and
can be used to instrument any custom python class.

Example: Instrumenting a custom class
    Consider a mock question-answering app with a context retriever component coded
    up as two classes in two python, `CustomApp` and `CustomRetriever`:

    ### `custom_app.py`

    ```python
    from trulens.apps.app import instrument
    from custom_retriever import CustomRetriever


    class CustomApp:
        # NOTE: No restriction on this class.

        def __init__(self):
            self.retriever = CustomRetriever()

        @instrument
        def retrieve_chunks(self, data):
            return self.retriever.retrieve_chunks(data)

        @instrument
        def respond_to_query(self, input):
            chunks = self.retrieve_chunks(input) output = f"The answer to {input} is
            probably {chunks[0]} or something ..." return output
    ```

    ### `custom_retriever.py`

    ```python
    from trulens.apps.app import instrument

    class CustomRetriever:
        # NOTE: No restriction on this class either.

        @instrument
        def retrieve_chunks(self, data):
            return [
                f"Relevant chunk: {data.upper()}", f"Relevant chunk: {data[::-1]}"
            ]
    ```

The core tool for instrumenting these classes is the `@instrument` decorator.
_TruLens_ needs to be aware of two high-level concepts to usefully monitor the
app: components and methods used by components. The `instrument` must decorate
each method that the user wishes to track.

The owner classes of any decorated method is then viewed as an app component. In
this example, case `CustomApp` and `CustomRetriever` are components.

Example:
    ### `example.py`

    ```python
    from custom_app import CustomApp
    from trulens.apps.app import TruApp

    custom_app = CustomApp()

    # Normal app Usage:
    response = custom_app.respond_to_query("What is the capital of Indonesia?")

    # Wrapping app with `TruApp`:
    tru_recorder = TruApp(ca)

    # Tracked usage:
    with tru_recorder:
        custom_app.respond_to_query, input="What is the capital of Indonesia?")
    ```

    `TruApp` constructor arguments are like in those higher-level
apps as well including the feedback functions, metadata, etc.

## Instrumenting 3rd party classes

In cases you do not have access to a class to make the necessary decorations for
tracking, you can instead use one of the static methods of `instrument`, for
example, the alternative for making sure the custom retriever gets instrumented
is via:

Example:
    ```python
    # custom_app.py`:

    from trulens.apps.app import instrument
    from some_package.from custom_retriever import CustomRetriever

    instrument.method(CustomRetriever, "retrieve_chunks")

    # ... rest of the custom class follows ...
    ```

## API Usage Tracking

Uses of python libraries for common LLMs like OpenAI are tracked in custom class
apps.

### Covered LLM Libraries

- Official OpenAI python package (https://github.com/openai/openai-python).
- Snowflake Cortex (https://docs.snowflake.com/en/sql-reference/functions/complete-snowflake-cortex.html).
- Amazon Bedrock (https://docs.aws.amazon.com/code-library/latest/ug/python_3_bedrock_code_examples.html).

### Huggingface

Uses of huggingface inference APIs are tracked as long as requests are made
through the `requests` class's `post` method to the URL
https://api-inference.huggingface.co .

## Limitations

- Tracked (instrumented) components must be accessible through other tracked
  components. Specifically, an app cannot have a custom class that is not
  instrumented but that contains an instrumented class. The inner instrumented
  class will not be found by trulens.

- All tracked components are categorized as "Custom" (as opposed to Template,
  LLM, etc.). That is, there is no categorization available for custom
  components. They will all show up as "uncategorized" in the dashboard.

- Non json-like contents of components (that themselves are not components) are
  not recorded or available in dashboard. This can be alleviated to some extent
  with the `app_extra_json` argument to `TruCustomClass` as it allows one to
  specify in the form of json additional information to store alongside the
  component hierarchy. Json-like (json bases like string, int, and containers
  like sequences and dicts are included).

## What can go wrong

- If a `with_record` or `awith_record` call does not encounter any instrumented
  method, it will raise an error. You can check which methods are instrumented
  using `App.print_instrumented`. You may have forgotten to decorate relevant
  methods with `@instrument`.

```python
app.print_instrumented()

### output example:
Components:
        TruApp (Other) at 0x171bd3380 with path *.__app__
        CustomApp (Custom) at 0x12114b820 with path *.__app__.app
        CustomLLM (Custom) at 0x12114be50 with path *.__app__.app.llm
        CustomMemory (Custom) at 0x12114bf40 with path *.__app__.app.memory
        CustomRetriever (Custom) at 0x12114bd60 with path *.__app__.app.retriever
        CustomTemplate (Custom) at 0x12114bf10 with path *.__app__.app.template

Methods:
Object at 0x12114b820:
        <function CustomApp.retrieve_chunks at 0x299132ca0> with path *.__app__.app
        <function CustomApp.respond_to_query at 0x299132d30> with path *.__app__.app
        <function CustomApp.arespond_to_query at 0x299132dc0> with path *.__app__.app
Object at 0x12114be50:
        <function CustomLLM.generate at 0x299106b80> with path *.__app__.app.llm
Object at 0x12114bf40:
        <function CustomMemory.remember at 0x299132670> with path *.__app__.app.memory
Object at 0x12114bd60:
        <function CustomRetriever.retrieve_chunks at 0x299132790> with path *.__app__.app.retriever
Object at 0x12114bf10:
        <function CustomTemplate.fill at 0x299132a60> with path *.__app__.app.template
```

- If an instrumented / decorated method's owner object cannot be found when
  traversing your custom class, you will get a warning. This may be ok in the
  end but may be indicative of a problem. Specifically, note the "Tracked"
  limitation above. You can also use the `app_extra_json` argument to `App` /
  `TruApp` to provide a structure to stand in place for (or augment) the
  data produced by walking over instrumented components to make sure this
  hierarchy contains the owner of each instrumented method.

  The owner-not-found error looks like this:

```python
Function <function CustomRetriever.retrieve_chunks at 0x177935d30> was not found during instrumentation walk. Make sure it is accessible by traversing app <custom_app.CustomApp object at 0x112a005b0> or provide a bound method for it as TruApp constructor argument `methods_to_instrument`.
Function <function CustomTemplate.fill at 0x1779474c0> was not found during instrumentation walk. Make sure it is accessible by traversing app <custom_app.CustomApp object at 0x112a005b0> or provide a bound method for it as TruApp constructor argument `methods_to_instrument`.
Function <function CustomLLM.generate at 0x1779471f0> was not found during instrumentation walk. Make sure it is accessible by traversing app <custom_app.CustomApp object at 0x112a005b0> or provide a bound method for it as TruApp constructor argument `methods_to_instrument`.
```

  Subsequent attempts at `with_record`/`awith_record` may result in the "Empty
  record" exception.

- Usage tracking not tracking. We presently have limited coverage over which
  APIs we track and make some assumptions with regards to accessible APIs
  through lower-level interfaces. Specifically, we only instrument the
  `requests` module's `post` method for the lower level tracking. Please file an
  issue on github with your use cases so we can work out a more complete
  solution as needed.
"""

import inspect
import logging
from pprint import PrettyPrinter
from typing import Any, Callable, ClassVar, Optional, Set

import pydantic
from pydantic import Field
from trulens.core import app as core_app
from trulens.core import experimental as core_experimental
from trulens.core import instruments as core_instruments
from trulens.core.session import TruSession
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import text as text_utils

logger = logging.getLogger(__name__)

pp = PrettyPrinter()

# Keys used in app_extra_json to indicate an automatically added structure for
# places an instrumented method exists but no instrumented data exists
# otherwise.
PLACEHOLDER = "__tru_placeholder"


class TruApp(core_app.App):
    """
    This recorder is the most flexible option for instrumenting an application,
    and can be used to instrument any custom python class.

    Track any custom app using methods decorated with `@instrument`, or whose
    methods are instrumented after the fact by `instrument.method`.

    Example: Using the `@instrument` decorator
        ```python
        from trulens.core import instrument

        class CustomApp:

            def __init__(self):
                self.retriever = CustomRetriever()
                self.llm = CustomLLM()
                self.template = CustomTemplate(
                    "The answer to {question} is probably {answer} or something ..."
                )

            @instrument
            def retrieve_chunks(self, data):
                return self.retriever.retrieve_chunks(data)

            @instrument
            def respond_to_query(self, input):
                chunks = self.retrieve_chunks(input)
                answer = self.llm.generate(",".join(chunks))
                output = self.template.fill(question=input, answer=answer)

                return output

        ca = CustomApp()
        ```

    Example: Using `instrument.method`
        ```python
        from trulens.core import instrument

        class CustomApp:

            def __init__(self):
                self.retriever = CustomRetriever()
                self.llm = CustomLLM()
                self.template = CustomTemplate(
                    "The answer to {question} is probably {answer} or something ..."
                )

            def retrieve_chunks(self, data):
                return self.retriever.retrieve_chunks(data)

            def respond_to_query(self, input):
                chunks = self.retrieve_chunks(input)
                answer = self.llm.generate(",".join(chunks))
                output = self.template.fill(question=input, answer=answer)

                return output

        custom_app = CustomApp()

        instrument.method(CustomApp, "retrieve_chunks")
        ```

    Once a method is tracked, its arguments and returns are available to be used
    in feedback functions. This is done by using the `Select` class to select
    the arguments and returns of the method.

    Doing so follows the structure:

    - For args: `Select.RecordCalls.<method_name>.args.<arg_name>`

    - For returns: `Select.RecordCalls.<method_name>.rets.<ret_name>`

    Example: "Defining feedback functions with instrumented methods"

        ```python
        f_context_relevance = (
            Feedback(provider.context_relevance_with_cot_reasons, name = "Context Relevance")
            .on(Select.RecordCalls.retrieve_chunks.args.query) # refers to the query arg of CustomApp's retrieve_chunks method
            .on(Select.RecordCalls.retrieve_chunks.rets.collect())
            .aggregate(np.mean)
            )
        ```

    Last, the `TruApp` recorder can wrap our custom application, and
    provide logging and evaluation upon its use.

    Example: Using the `TruApp` recorder
        ```python
        from trulens.apps.app import TruApp

        tru_recorder = TruApp(custom_app,
            app_name="Custom Application",
            app_version="base",
            feedbacks=[f_context_relevance])

        with tru_recorder as recording:
            custom_app.respond_to_query("What is the capital of Indonesia?")
        ```

        See [Feedback
        Functions](https://www.trulens.org/trulens/api/feedback/) for
        instantiating feedback functions.

    Args:
        app: Any class.

        **kwargs: Additional arguments to pass to [App][trulens.core.app.App]
            and [AppDefinition][trulens.core.schema.app.AppDefinition]
    """

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        arbitrary_types_allowed=True
    )

    app: Any

    root_callable: ClassVar[pyschema_utils.FunctionOrMethod] = Field(None)

    functions_to_instrument: ClassVar[Set[Callable]] = set()
    """Methods marked as needing instrumentation.

    These are checked to make sure the object walk finds them. If not, a message
    is shown to let user know how to let the TruApp constructor know where
    these methods are.
    """

    def __init__(
        self,
        app: Any,
        main_method: Optional[Callable] = None,
        methods_to_instrument=None,
        **kwargs: Any,
    ):
        kwargs["app"] = app
        tru_session = (
            TruSession()
            if "connector" not in kwargs
            else TruSession(connector=kwargs["connector"])
        )
        if tru_session.experimental_feature(
            core_experimental.Feature.OTEL_TRACING
        ):
            main_methods = set()
            if main_method is not None:
                main_methods.add(main_method)
            for _, method in inspect.getmembers(app, inspect.ismethod):
                if self._has_record_root_instrumentation(method):
                    main_methods.add(method)
            if len(main_methods) != 1:
                raise ValueError(
                    f"Must have exactly one main method or method decorated with span type 'record_root'! Found: {list(main_methods)}"
                )
            main_method = main_methods.pop()
        if main_method is not None:
            kwargs["main_method"] = main_method
        kwargs["root_class"] = pyschema_utils.Class.of_object(app)

        instrument = core_instruments.Instrument(
            app=self  # App mixes in WithInstrumentCallbacks
        )
        kwargs["instrument"] = instrument

        # This does instrumentation:
        super().__init__(**kwargs)

        methods_to_instrument = methods_to_instrument or dict()

        # The rest of this code instruments methods explicitly passed to
        # constructor as needing instrumentation and checks that methods
        # decorated with @instrument or passed explicitly belong to some
        # component as per serialized version of this app. If they are not,
        # placeholders are made in `app_extra_json` so that subsequent
        # serialization looks like the components exist.
        json = self.model_dump()

        for m, path in methods_to_instrument.items():
            method_name = m.__name__

            full_path = serial_utils.Lens().app + path

            self.instrument.instrument_method(
                method_name=method_name, obj=m.__self__, query=full_path
            )

            # TODO: DEDUP with next condition

            # Check whether the path/location of the method is in json serialization and
            # if not, add a placeholder to app_extra_json.
            try:
                next(full_path(json))

                print(
                    f"{text_utils.UNICODE_CHECK} Added method {m.__name__} under component at path {full_path}"
                )

            except Exception:
                logger.warning(
                    f"App has no component at path {full_path} . "
                    f"Specify the component with the `app_extra_json` argument to TruApp constructor. "
                    f"Creating a placeholder there for now."
                )

                path.set(
                    self.app_extra_json,
                    {
                        PLACEHOLDER: "I was automatically added to `app_extra_json` because there was nothing here to refer to an instrumented method owner.",
                        m.__name__: f"Placeholder for method {m.__name__}.",
                    },
                )

        # Check that any functions marked with `TruApp.instrument` has been
        # instrumented as a method under some object.
        for f in TruApp.functions_to_instrument:
            obj_ids_methods_and_full_paths = list(self.get_methods_for_func(f))

            if len(obj_ids_methods_and_full_paths) == 0:
                logger.warning(
                    f"Function {f} was not found during instrumentation walk. "
                    f"Make sure it is accessible by traversing app {app} "
                    f"or provide a bound method for it as TruApp constructor argument `methods_to_instrument`."
                )

            else:
                for obj_id, m, full_path in obj_ids_methods_and_full_paths:
                    try:
                        next(full_path.get(json))

                    except Exception:
                        logger.debug(
                            f"App has no component owner of instrumented method {m} at path {full_path}. "
                            f"Specify the component with the `app_extra_json` argument to TruApp constructor. "
                            f"Creating a placeholder there for now."
                        )

                        full_path.set(
                            self.app_extra_json,
                            {
                                PLACEHOLDER: "I was automatically added to `app_extra_json` because there was nothing here to refer to an instrumented method owner.",
                                m.__name__: f"Placeholder for method {m.__name__}.",
                            },
                        )

    def main_call(self, human: str):
        if self.main_method_name is None:
            raise RuntimeError(
                "`main_method_name` was not specified so we do not know how to run this app."
            )

        main_method = getattr(self.app, self.main_method_name)

        return main_method(human)


class instrument(core_instruments.instrument):
    """
    Decorator for marking methods to be instrumented in custom classes that are
    wrapped by TruApp.
    """

    @classmethod
    def method(cls, inst_cls: type, name: str) -> None:
        core_instruments.instrument.method(inst_cls, name)

        # Also make note of it for verification that it was found by the walk
        # after init.
        TruApp.functions_to_instrument.add(getattr(inst_cls, name))


TruApp.model_rebuild()



"""Utilities related to core python functionalities."""

from __future__ import annotations

import asyncio
from collections import defaultdict
from contextlib import asynccontextmanager
from contextlib import contextmanager
import contextvars
import dataclasses
import inspect
import logging
from pprint import PrettyPrinter
import queue
import sys
import threading
from types import FrameType
from types import ModuleType
import typing
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)
import weakref

import pydantic
from trulens.core._utils.pycompat import ReferenceType

T = TypeVar("T")

WRAP_LAZY: bool = True

Thunk = Callable[[], T]
"""A function that takes no arguments."""

logger = logging.getLogger(__name__)
pp = PrettyPrinter()

# Reflection utilities.


def class_name(obj: Union[Type, Any]) -> str:
    """Get the class name of the given object or instance."""

    if hasattr(obj, "__name__"):
        return obj.__name__

    if hasattr(obj, "__class__"):
        return obj.__class__.__name__

    return str(obj)


def module_name(obj: Union[ModuleType, Type, Any]) -> str:
    """Get the module name of the given module, class, or instance."""

    if isinstance(obj, ModuleType):
        return obj.__name__

    if hasattr(obj, "__module__"):
        return obj.__module__  # already a string name

    return "unknown module"


def callable_name(c: Callable):
    """Get the name of the given callable."""

    if isinstance(c, staticmethod):
        return callable_name(c.__func__)

    if isinstance(c, classmethod):
        return callable_name(c.__func__)

    if not isinstance(c, Callable):
        raise ValueError(
            f"Expected a callable. Got {class_name(type(c))} instead."
        )

    if safe_hasattr(c, "__name__"):
        return c.__name__

    if safe_hasattr(c, "__call__"):
        return callable_name(c.__call__)

    return str(c)


def id_str(obj: Any) -> str:
    """Get the id of the given object as a string in hex."""

    return f"0x{id(obj):x}"


def is_really_coroutinefunction(func) -> bool:
    """Determine whether the given function is a coroutine function.

    Warning:
        Inspect checkers for async functions do not work on openai clients,
        perhaps because they use `@typing.overload`. Because of that, we detect
        them by checking `__wrapped__` attribute instead. Note that the inspect
        docs suggest they should be able to handle wrapped functions but perhaps
        they handle different type of wrapping? See
        https://docs.python.org/3/library/inspect.html#inspect.iscoroutinefunction
        . Another place they do not work is the decorator langchain uses to mark
        deprecated functions.
    """

    if inspect.iscoroutinefunction(func):
        return True

    if hasattr(func, "__wrapped__") and inspect.iscoroutinefunction(
        func.__wrapped__
    ):
        return True

    return False


def safe_signature(func_or_obj: Any):
    """Get the signature of the given function.

    Sometimes signature fails for wrapped callables and in those cases we check
    for `__call__` attribute and use that instead.
    """
    try:
        assert isinstance(
            func_or_obj, Callable
        ), f"Expected a Callable. Got {type(func_or_obj)} instead."

        return inspect.signature(func_or_obj)

    except Exception as e:
        if safe_hasattr(func_or_obj, "__call__"):
            # If given an obj that is callable (has __call__ defined), we want to
            # return signature of that call instead of letting inspect.signature
            # explore that object further. Doing so may produce exceptions due to
            # contents of those objects producing exceptions when attempting to
            # retrieve them.

            return inspect.signature(func_or_obj.__call__)

        else:
            raise e


def safer_getattr(obj: Any, k: str, default: Optional[Any] = None) -> Any:
    """Get the attribute `k` of the given object.

    Returns default if the attribute cannot be retrieved.
    """

    try:
        return safe_getattr(obj, k)
    except Exception:
        return default


def safe_getattr(obj: Any, k: str, get_prop: bool = True) -> Any:
    """Try to get the attribute `k` of the given object.

    This may evaluate some code if the attribute is a property and may fail. If
    `get_prop` is False, will not return contents of properties (will raise
    `ValueException`).
    """

    v = inspect.getattr_static(obj, k)

    is_prop = False
    try:
        # OpenAI version 1 classes may cause this isinstance test to raise an
        # exception.
        is_prop = isinstance(v, property)
    except Exception as e:
        raise RuntimeError(f"Failed to check if {k} is a property.") from e

    if is_prop:
        if not get_prop:
            raise ValueError(f"{k} is a property")

        if v.fget is None:
            raise ValueError(f"{k} property does not have a getter.")

        try:
            v = v.fget(obj)
            return v

        except Exception as e:
            raise RuntimeError(f"Failed to get property {k}.") from e

    else:
        return v


def safe_hasattr(obj: Any, k: str) -> bool:
    """Check if the given object has the given attribute.

    Attempts to use static checks (see [inspect.getattr_static][]) to avoid any
    side effects of attribute access (i.e. for properties).
    """
    try:
        v = inspect.getattr_static(obj, k)
    except AttributeError:
        return False

    is_prop = False
    try:
        # OpenAI version 1 classes may cause this isinstance test to raise an
        # exception.
        is_prop = isinstance(v, property)
    except Exception:
        return False

    if is_prop:
        try:
            v.fget(obj)
            return True
        except Exception:
            return False
    else:
        return True


def safe_issubclass(cls: Type, parent: Type) -> bool:
    """Check if the given class is a subclass of the given parent class."""

    origin = typing.get_origin(cls)
    if origin is None:
        return issubclass(cls, parent)

    return issubclass(origin, parent)


# Function utilities.


def code_line(func, show_source: bool = False) -> Optional[str]:
    """Get a string representation of the location of the given function
    `func`."""

    if isinstance(func, inspect.FrameInfo):
        ret = f"{func.filename}:{func.lineno}"
        if show_source and func.code_context is not None:
            ret += "\n"
            for line in func.code_context:
                ret += "\t" + line

        return ret

    if inspect.isframe(func):
        code = func.f_code
        ret = f"{func.f_code.co_filename}:{func.f_code.co_firstlineno}"

    elif safe_hasattr(func, "__code__"):
        code = func.__code__
        ret = f"{code.co_filename}:{code.co_firstlineno}"

    else:
        return None

    if show_source:
        try:
            ret += "\n"
            for line in inspect.getsourcelines(func)[0]:
                ret += "\t" + str(line)
        except OSError as e:
            ret += f"Source code cannot be retrieved: {e}"

    return ret


def locals_except(*exceptions):
    """
    Get caller's locals except for the named exceptions.
    """

    locs = caller_frame(offset=1).f_locals  # 1 to skip this call

    return {k: v for k, v in locs.items() if k not in exceptions}


def for_all_methods(decorator, _except: Optional[List[str]] = None):
    """
    Applies decorator to all methods except classmethods, private methods and
    the ones specified with `_except`.
    """

    def decorate(cls):
        for (
            attr_name,
            attr,
        ) in cls.__dict__.items():  # does not include classmethods
            if not inspect.isfunction(attr):
                continue  # skips non-method attributes

            if attr_name.startswith("_"):
                continue  # skips private methods

            if _except is not None and attr_name in _except:
                continue

            logger.debug("Decorating %s", attr_name)
            setattr(cls, attr_name, decorator(attr))

        return cls

    return decorate


def run_before(callback: Callable):
    """
    Create decorator to run the callback before the function.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            callback(*args, **kwargs)
            return func(*args, **kwargs)

        return wrapper

    return decorator


# Python call stack utilities

# Attribute name for storing a callstack in asyncio tasks.
STACK = "__tru_stack"


def superstack() -> Iterator[FrameType]:
    """Get the current stack (not including this function) with frames reaching
    across Tasks and threads.
    """

    frames = stack_with_tasks()
    next(iter(frames))  # skip this method itself
    # NOTE: skipping offset frames is done below since the full stack may need
    # to be reconstructed there.

    # Using queue for frames as additional frames may be added due to handling threads.
    q = queue.Queue()
    for f in frames:
        q.put(f)

    while not q.empty():
        f = q.get()
        yield f

        if id(f.f_code) == id(_future_target_wrapper.__code__):
            locs = f.f_locals

            assert (
                "pre_start_stack" in locs
            ), "Pre thread start stack expected but not found."
            for fi in locs["pre_start_stack"].get():  # is WeakWrapper
                q.put(fi.frame)

            continue

    return


def caller_module_name(offset=0) -> str:
    """
    Get the caller's (of this function) module name.
    """
    frame = caller_frame(offset=offset + 1)
    return frame.f_globals["__name__"]


def caller_module(offset=0) -> ModuleType:
    """
    Get the caller's (of this function) module.
    """

    return sys.modules[caller_module_name(offset=offset + 1)]


def caller_frame(offset=0) -> FrameType:
    """
    Get the caller's (of this function) frame. See
    https://docs.python.org/3/reference/datamodel.html#frame-objects .
    """
    caller_frame = inspect.currentframe()
    for _ in range(offset + 1):
        if caller_frame is None:
            raise RuntimeError("No current frame found.")
        caller_frame = caller_frame.f_back

    if caller_frame is None:
        raise RuntimeError("No caller frame found.")

    return caller_frame


def external_caller_frame(offset=0) -> FrameType:
    """Get the caller's (of this function) frame that is not in the trulens
    namespace.

    Raises:
        RuntimeError: If no such frame is found.
    """
    frame = inspect.currentframe()
    gen = stack_generator(frame=frame, offset=offset + 1)
    for f_info in gen:
        name = f_info.f_globals["__name__"]
        if not (name.startswith("trulens") or name.startswith("pydantic")):
            return f_info

    raise RuntimeError("No external caller frame found.")


def caller_frameinfo(
    offset: int = 0, skip_module: Optional[str] = "trulens"
) -> Optional[inspect.FrameInfo]:
    """
    Get the caller's (of this function) frameinfo. See
    https://docs.python.org/3/reference/datamodel.html#frame-objects .

    Args:
        offset: The number of frames to skip. Default is 0.

        skip_module: Skip frames from the given module. Default is "trulens".
    """

    for f_info in inspect.stack(0)[offset + 1 :]:
        if skip_module is None:
            return f_info
        if not f_info.frame.f_globals["__name__"].startswith(skip_module):
            return f_info

    return None


def task_factory_with_stack(loop, coro, *args, **kwargs) -> asyncio.Task:
    """A task factory that annotates created tasks with stacks and context of
    their parents.

    All of such annotated stacks can be retrieved with
    [stack_with_tasks][trulens.core.utils.python.stack_with_tasks] as one merged
    stack.
    """

    if "context" in kwargs:
        logger.warning(
            "Context is being overwritten, TruLens may not be able to record traces."
        )

    parent_task = asyncio.current_task(loop=loop)
    task = asyncio.tasks.Task(coro=coro, loop=loop, *args, **kwargs)

    frame = inspect.currentframe()
    stack = stack_generator(frame=frame, offset=3)

    if parent_task is not None:
        stack = merge_stacks(stack, parent_task.get_stack()[::-1])
        # skipping create_task and task_factory

    setattr(task, STACK, stack)

    return task


# If there is already a loop running, try to patch its task factory.
try:
    loop = asyncio.get_running_loop()
    loop.set_task_factory(task_factory_with_stack)
except Exception:
    pass

# Instrument new_event_loop to set the above task_factory upon creation:
original_new_event_loop = asyncio.new_event_loop


def tru_new_event_loop():
    """Replacement for [new_event_loop][asyncio.new_event_loop] that sets
    the task factory to make tasks that copy the stack from their creators."""

    loop = original_new_event_loop()
    loop.set_task_factory(task_factory_with_stack)
    return loop


asyncio.new_event_loop = tru_new_event_loop


def get_task_stack(task: asyncio.Task) -> Sequence[FrameType]:
    """Get the annotated stack (if available) on the given task."""

    if safe_hasattr(task, STACK):
        return getattr(task, STACK)
    else:
        # get_stack order is reverse of inspect.stack:
        return task.get_stack()[::-1]


def merge_stacks(
    s1: Iterable[FrameType], s2: Sequence[FrameType]
) -> Sequence[FrameType]:
    """
    Assuming `s1` is a subset of `s2`, combine the two stacks in presumed call
    order.
    """

    ret = []

    for f in s1:
        ret.append(f)
        try:
            s2i = s2.index(f)
            for _ in range(s2i):
                ret.append(s2[0])
                s2 = s2[1:]
        except Exception:
            pass

    return ret


def stack_generator(
    frame: Optional[FrameType] = None, offset: int = 0
) -> Iterable[FrameType]:
    if frame is None:
        frame = inspect.currentframe()
    for _ in range(offset):
        if frame is None:
            raise ValueError("No frame found.")
        frame = frame.f_back
    while frame is not None:
        yield frame
        frame = frame.f_back


def stack_with_tasks() -> Iterable[FrameType]:
    """Get the current stack (not including this function) with frames reaching
    across Tasks.
    """
    frame = inspect.currentframe()
    frame_gen = stack_generator(frame=frame, offset=1)
    try:
        task_stack = get_task_stack(asyncio.current_task())

        return merge_stacks(frame_gen, task_stack)

    except Exception:
        return frame_gen


class _Wrap(Generic[T]):
    """Wrap an object.

    See WeakWrapper for explanation.
    """

    def __init__(self, obj: T):
        self.obj: T = obj

    def get(self) -> T:
        return self.obj


@dataclasses.dataclass
class WeakWrapper(Generic[T]):
    """Wrap an object with a weak reference.

    This is to be able to use weakref.ref on objects like lists which are
    otherwise not weakly referenceable. The goal of this class is to generalize
    weakref.ref to work with any object."""

    obj: ReferenceType[Union[_Wrap[T], T]]

    def __init__(self, obj: Union[ReferenceType[T], WeakWrapper[T], T]):
        if isinstance(obj, ReferenceType):
            self.obj = obj

        else:
            if isinstance(obj, WeakWrapper):
                obj = obj.get()

            try:
                # Try to make reference to obj directly.
                self.obj = weakref.ref(obj)

            except Exception:
                # If its a list or other non-weakly referenceable object, wrap it.
                self.obj = weakref.ref(_Wrap(obj))

    def get(self) -> T:
        """Get the wrapped object."""

        temp = self.obj()  # undo weakref.ref
        if isinstance(temp, _Wrap):
            return temp.get()  # undo _Wrap if needed
        else:
            return temp


def _future_target_wrapper(stack, func, *args, **kwargs):
    """Wrapper for a function that is started by threads.

    This is needed to record the call stack prior to thread creation as in
    python threads do not inherit the stack. Our instrumentation, however,
    relies on walking the stack and need to do this to the frames prior to
    thread starts.
    """

    # Keep this for looking up via get_first_local_in_call_stack .
    pre_start_stack = WeakWrapper(stack)  # noqa: F841 # pylint: disable=W0612

    # with with_context(context):
    return func(*args, **kwargs)


def get_all_local_in_call_stack(
    key: str,
    func: Callable[[Callable], bool],
    offset: Optional[int] = 1,
    skip: Optional[Any] = None,  # really frame
) -> Iterator[Any]:
    """Find locals in call stack by name.

    Args:
        key: The name of the local variable to look for.

        func: Recognizer of the function to find in the call stack.

        offset: The number of top frames to skip.

        skip: A frame to skip as well.

    Note:
        `offset` is unreliable for skipping the intended frame when operating
        with async tasks. In those cases, the `skip` argument is more reliable.

    Returns:
        An iterator over the values of the local variable named `key` in the
            stack at all of the frames executing a function which `func` recognizes
            (returns True on) starting from the top of the stack except `offset` top
            frames.

            Returns None if `func` does not recognize any function in the stack.

    Raises:
        RuntimeError: Raised if a function is recognized but does not have `key`
            in its locals.

    This method works across threads as long as they are started using
    [TP][trulens.core.utils.threading.TP].
    """

    frames_gen = stack_with_tasks()
    # NOTE: skipping offset frames is done below since the full stack may need
    # to be reconstructed there.

    # Using queue for frames as additional frames may be added due to handling threads.
    q = queue.Queue()
    for i, f in enumerate(frames_gen):
        if i == 0:
            # skip this method itself
            continue
        q.put(f)

    while not q.empty():
        f = q.get()

        if id(f.f_code) == id(_future_target_wrapper.__code__):
            locs = f.f_locals
            assert (
                "pre_start_stack" in locs
            ), "Pre thread start stack expected but not found."
            for fi in locs["pre_start_stack"].get():  # is WeakWrapper
                q.put(fi.frame)

            continue

        if offset is not None and offset > 0:
            offset -= 1
            continue

        if func(f.f_code):
            logger.debug(
                "Looking via %s; found %s", callable_name(func), str(f)
            )
            if skip is not None and f == skip:
                logger.debug("Skipping.")
                continue

            locs = f.f_locals
            if key in locs:
                yield locs[key]
            else:
                raise KeyError(f"No local named '{key}' found in frame {f}.")

    return


def get_first_local_in_call_stack(
    key: str,
    func: Callable[[Callable], bool],
    offset: Optional[int] = 1,
    skip: Optional[Any] = None,  # actually frame
) -> Optional[Any]:
    """
    Get the value of the local variable named `key` in the stack at the nearest
    frame executing a function which `func` recognizes (returns True on)
    starting from the top of the stack except `offset` top frames. If `skip`
    frame is provided, it is skipped as well. Returns None if `func` does not
    recognize the correct function. Raises RuntimeError if a function is
    recognized but does not have `key` in its locals.

    This method works across threads as long as they are started using the TP
    class above.

    NOTE: `offset` is unreliable for skipping the intended frame when operating
    with async tasks. In those cases, the `skip` argument is more reliable.
    """

    try:
        return next(
            iter(
                get_all_local_in_call_stack(
                    key, func, offset=offset + 1, skip=skip
                )
            )
        )
    except StopIteration:
        logger.debug("no frames found")
        return None


# Wrapping utilities


class OpaqueWrapper(Generic[T]):
    """Wrap an object preventing all access.

    Any access except to
    [unwrap][trulens.core.utils.python.OpaqueWrapper.unwrap] will result in an
    exception with the given message.

    Args:
        obj: The object to wrap.

        e: The exception to raise when an attribute is accessed.
    """

    def __init__(self, obj: T, e: Exception):
        self._obj = obj
        self._e = e

    def unwrap(self) -> T:
        """Get the wrapped object back."""
        return self._obj

    def __getattr__(self, name):
        raise self._e

    def __setattr__(self, name, value):
        if name in ["_obj", "_e"]:
            return super().__setattr__(name, value)
        raise self._e

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        raise self._e


ContextVarsOrValues = Union[
    Iterable[contextvars.ContextVar],
    contextvars.Context,
    Dict[contextvars.ContextVar, Any],
]


def set_context_vars_or_values(
    context_vars: Optional[ContextVarsOrValues] = None,
) -> Dict[contextvars.ContextVar, contextvars.Token]:
    """Get the tokens for the given context variables or values.

    Args:
        context_vars: The context variables or values to get tokens for.

    Returns:
        A dictionary of context variables to tokens.
    """

    if context_vars is None:
        return {}

    if isinstance(context_vars, (dict, contextvars.Context)):
        return {cv: cv.set(v) for cv, v in context_vars.items()}

    return {cv: cv.set(cv.get()) for cv in context_vars}


@contextmanager
def with_context(context_vars: Optional[ContextVarsOrValues] = None):
    """Context manager to set context variables to given values.

    Args:
        context_vars: The context variables to set. If a dictionary is given,
            the keys are the context variables and the values are the values to
            set them to. If an iterable is given, it should be a list of context
            variables to set to their current value.
    """

    try:
        tokens = set_context_vars_or_values(context_vars)
        yield

    finally:
        for cv, v in tokens.items():
            try:
                cv.reset(v)
            except Exception as e:
                logger.warning("Context reset failed: %s", e)


@asynccontextmanager
async def awith_context(context_vars: Optional[ContextVarsOrValues] = None):
    """Context manager to set context variables to given values.

    Args:
        context_vars: The context variables to set. If a dictionary is given,
            the keys are the context variables and the values are the values to
            set them to. If an iterable is given, it should be a list of context
            variables to set to their current value.
    """

    try:
        tokens = set_context_vars_or_values(context_vars)
        yield

    finally:
        for cv, v in tokens.items():
            try:
                cv.reset(v)
            except Exception as e:
                logger.warning("Context reset failed: %s", e)


def wrap_awaitable(
    awaitable: Awaitable[T],
    on_await: Optional[Callable[[], Any]] = None,
    wrap: Optional[Callable[[T], T]] = None,
    on_done: Optional[Callable[[T], T]] = None,
    context_vars: Optional[ContextVarsOrValues] = None,
) -> Awaitable[T]:
    """Wrap an awaitable in another awaitable that will call callbacks before
    and after the given awaitable finishes.

    !!! Important
        This method captures a [Context][contextvars.Context] at the time this
        method is called and copies it over to the wrapped awaitable.

    Note that the resulting awaitable needs to be awaited for the callback to
    eventually trigger.

    Args:
        awaitable: The awaitable to wrap.

        on_await: The callback to call when the wrapper awaitable is awaited but
            before the wrapped awaitable is awaited.

        wrap: The callback to call with the result of the wrapped awaitable
            once it is ready. This should return the value or a wrapped version.

        on_done: For compatibility with generators, this is called after wrap.

        context_vars: The context variables to copy over to the wrapped
            awaitable. If None, all context variables are copied. See
            [with_context][trulens.core.utils.python.with_context].
    """

    async def wrapper(awaitable):
        async with awith_context(context_vars):
            if on_await is not None:
                on_await()

            val = await awaitable

            if wrap is not None:
                val = wrap(val)  # allow handlers to transform the value

            if on_done is not None:
                val = on_done(val)

            return val

    wrapper.__name__ = f"tru_wrapped_{awaitable.__class__.__name__}"

    return wrapper(awaitable)


def wrap_generator(
    gen: Generator[T, None, None],
    on_iter: Optional[Callable[[], Any]] = None,
    wrap: Optional[Callable[[T], T]] = None,
    on_done: Optional[Callable[[List[T]], Any]] = None,
    context_vars: Optional[ContextVarsOrValues] = None,
) -> Generator[T, None, None]:
    """Wrap a generator in another generator that will call callbacks at various
    points in the generation process.

    Args:
        gen: The generator to wrap.

        on_iter: The callback to call when the wrapper generator is created but
            before a first iteration is produced.

        wrap: The callback to call with the result of each iteration of the
            wrapped generator. This should return the value or a wrapped
            version.

        on_done: The callback to call when the wrapped generator is exhausted.

        context_vars: The context variables to copy over to the wrapped
            generator. If None, all context variables are taken with their
            present values. See
            [with_context][trulens.core.utils.python.with_context].
    """

    def wrapper(gen):
        with with_context(context_vars):
            if on_iter is not None:
                on_iter()

            vals = []

            for val in gen:
                if wrap is not None:
                    # Allow handlers to transform the value.
                    val = wrap(val)

                vals.append(val)
                yield val

            if on_done is not None:
                on_done(vals)

    wrapper.__name__ = f"tru_wrapped_{gen.__class__.__name__}"

    return wrapper(gen)


def wrap_async_generator(
    gen: AsyncGenerator[T, None],
    on_iter: Optional[Callable[[], Any]] = None,
    wrap: Optional[Callable[[T], T]] = None,
    on_done: Optional[Callable[[List[T]], Any]] = None,
    context_vars: Optional[ContextVarsOrValues] = None,
) -> AsyncGenerator[T, None]:
    """Wrap a generator in another generator that will call callbacks at various
    points in the generation process.

    Args:
        gen: The generator to wrap.

        on_iter: The callback to call when the wrapper generator is created but
            before a first iteration is produced.

        wrap: The callback to call with the result of each iteration of the
            wrapped generator.

        on_done: The callback to call when the wrapped generator is exhausted.

        context_vars: The context variables to copy over to the wrapped
            generator. If None, all context variables are taken with their
            present values. See
            [with_context][trulens.core.utils.python.with_context].
    """

    async def wrapper(gen):
        with with_context(context_vars):
            if on_iter is not None:
                on_iter()

            vals = []

            async for val in gen:
                if wrap is not None:
                    val = wrap(val)  # allow handlers to rewrap the value

                vals.append(val)

                yield val

            if on_done is not None:
                on_done(vals)

    wrapper.__name__ = f"tru_wrapped_{gen.__class__.__name__}"

    return wrapper(gen)


def is_lazy(obj):
    """Check if the given object is lazy.

    An object is considered lazy if it is a generator or an awaitable.
    """

    return (
        inspect.isawaitable(obj)
        or inspect.isgenerator(obj)
        or inspect.isasyncgen(obj)
    )


def wrap_lazy(
    obj: Any,
    on_start: Optional[Callable[[], None]] = None,
    wrap: Optional[Callable[[T], T]] = None,
    on_done: Optional[Callable[[Any], Any]] = None,  # Any may be T or List[T]
    context_vars: Optional[ContextVarsOrValues] = None,
) -> Any:
    """Wrap a lazy value in one that will call
    callbacks at various points in the generation process.

    Args:
        obj: The lazy value.

        on_start: The callback to call when the wrapper is created.

        wrap: The callback to call with the result of each iteration of the
            wrapped generator or the result of an awaitable. This should return
            the value or a wrapped version.

        on_done: The callback to call when the wrapped generator is exhausted or
            awaitable is ready.

        context_vars: The context variables to copy over to the wrapped
            generator. If None, all context variables are taken with their
            present values. See
            [with_context][trulens.core.utils.python.with_context].
    """

    if not WRAP_LAZY:
        return obj

    if inspect.isasyncgen(obj):
        return wrap_async_generator(
            obj,
            on_iter=on_start,
            wrap=wrap,
            on_done=on_done,
            context_vars=context_vars,
        )

    if inspect.isgenerator(obj):
        return wrap_generator(
            obj,
            on_iter=on_start,
            wrap=wrap,
            on_done=on_done,
            context_vars=context_vars,
        )

    if inspect.isawaitable(obj):
        return wrap_awaitable(
            obj,
            on_await=on_start,
            wrap=wrap,
            on_done=on_done,
            context_vars=context_vars,
        )

    raise ValueError(f"Object of type {type(obj)} is not lazy.")


def wrap_until_eager(
    obj,
    on_eager: Optional[Callable[[Any], T]] = None,
    context_vars: Optional[ContextVarsOrValues] = None,
) -> T | Sequence[T]:
    """Wrap a lazy value in one that will call callbacks one the final non-lazy
    values.

    Arts:
        obj: The lazy value.

        on_eager: The callback to call with the final value of the wrapped
            generator or the result of an awaitable. This should return the
            value or a wrapped version.

        context_vars: The context variables to copy over to the wrapped
            generator. If None, all context variables are taken with their
            present values. See
            [with_context][trulens.core.utils.python.with_context].
    """

    def rewrap(obj_):
        if is_lazy(obj_):
            return wrap_lazy(
                obj_, wrap=rewrap, on_done=rewrap, context_vars=context_vars
            )

        if on_eager is not None:
            return on_eager(obj_)

        return obj_

    return rewrap(obj)


# Context utilities


def context_id() -> str:
    """Return a short representation of context that includes the thread and
    async task identifiers."""

    ret = "T[" + threading.current_thread().name + "]"
    try:
        task = asyncio.current_task()
        if task is not None:
            ret += " K[" + task.name + "]"
    except Exception:
        pass

    return ret


# Class utilities


class SingletonPerNameMeta(type):
    """Metaclass for creating singleton instances except there being one
    instance max, there is one max per different `name` argument.

    If `name` is never given, reverts to normal singleton behavior.
    """

    _singleton_instances = {}

    def __call__(cls, *args, name: Optional[str] = None, **kwargs):
        """
        Create the singleton instance if it doesn't already exist and return it.
        """
        k = f"{cls.__module__}.{cls.__name__}", name

        if k not in cls._singleton_instances:
            logger.debug(
                "Creating new %s singleton instance for name = %s.",
                cls.__name__,
                name,
            )
            cls._singleton_instances[k] = super(
                SingletonPerNameMeta, cls
            ).__call__(*args, **kwargs)
        elif args or kwargs:
            logger.warning(
                "Singleton instance %s already exists for name = %s.",
                cls.__name__,
                name,
            )
        return cls._singleton_instances[k]

    @staticmethod
    def delete_singleton_by_name(
        name: str,
        cls: Optional[Type[Any]] = None,  # Any is the type with this as meta
    ):
        """Delete the singleton instance with the given name.

        This can be used for testing to create another singleton.

        Args:
            name: The name of the singleton instance to delete.

            cls: The class of the singleton instance to delete. If not given, all
                instances with the given name are deleted.
        """

        for k, v in list(SingletonPerNameMeta._singleton_instances.items()):
            if k[1] == name:
                if cls is not None and v.__class__ is not cls:
                    continue

                del SingletonPerNameMeta._singleton_instances[k]

    @staticmethod
    def delete_singleton(
        obj: Type[SingletonPerNameMeta], name: Optional[str] = None
    ):
        """
        Delete the singleton instance. Can be used for testing to create another
        singleton.
        """
        cls_name = (
            getattr(obj.__class__, "__name__")
            if hasattr(obj.__class__, "__name__")
            else None
        )
        k = cls_name, name
        if k in SingletonPerNameMeta._singleton_instances:
            del SingletonPerNameMeta._singleton_instances[k]
        else:
            logger.warning("Instance %s not found:", obj)


class PydanticSingletonMeta(type(pydantic.BaseModel), SingletonPerNameMeta):
    """This is the metaclass for creating Pydantic models that are also required
    to be singletons"""

    pass


class PydanticSingleton(metaclass=PydanticSingletonMeta):
    """A Pydantic model that is also a singleton by name.

    This only specifies the metaclass and wraps `delete_singleton_by_name`.
    """

    @classmethod
    def delete_singleton_by_name(cls, name: str):
        SingletonPerNameMeta.delete_singleton_by_name(name, cls)


class InstanceRefMixin:
    """Mixin for classes that need to keep track of their instances."""

    _instance_refs: Dict[Type, List[ReferenceType[InstanceRefMixin]]] = (
        defaultdict(list)
    )

    def __init__(self, register_instance: bool = True):
        if register_instance:
            self._instance_refs[self.__class__].append(weakref.ref(self))

    @classmethod
    def get_instances(cls) -> Generator[InstanceRefMixin]:
        """Get all instances of the class."""
        for inst_ref in cls._instance_refs[cls]:
            inst = inst_ref()
            if inst is not None:
                yield inst

    @classmethod
    def delete_instances(cls):
        """Delete all instances of the class."""

        cls._instance_refs[cls] = []



"""
Soon to be deprecated in favor of [TruApp][trulens.apps.app.TruApp].
"""

import logging
from pprint import PrettyPrinter
import warnings

from trulens.apps.app import TruApp
from trulens.apps.app import instrument as new_instrument

logger = logging.getLogger(__name__)

pp = PrettyPrinter()

# Keys used in app_extra_json to indicate an automatically added structure for
# places an instrumented method exists but no instrumented data exists
# otherwise.
PLACEHOLDER = "__tru_placeholder"


class TruCustomApp(TruApp):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "TruCustomApp is being deprecated in the next major version; use TruApp instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)


warnings.warn(
    """from trulens.apps.custom import instrument
        is being deprecated in the next major version; use from trulens.apps.app import instrument
        instead.""",
    DeprecationWarning,
    stacklevel=2,
)

# Alias instrument to the new location while keeping the old import path working
instrument = new_instrument


TruCustomApp.model_rebuild()
